import os

import pytest
from pytest import approx
import pandas as pd


import calliope
from calliope.core.attrdict import AttrDict
from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import check_error_or_warning


def get_supply_conversion_techs(model):
    available_techs = [
        "elec_to_heat_cool_unlinked",
        "cheap_elec_supply",
        "elec_to_heat_cool_linked",
        "elec_to_heat",
        "expensive_elec_supply",
        "normal_elec_supply",
        "cheap_heat_supply",
        "cheap_cool_supply",
        "normal_heat_supply",
        "expensive_heat_supply",
        "normal_cool_supply",
        "expensive_cool_supply",
        "elec_supply_plus",
    ]
    return [i for i in available_techs if i in model._model_data.techs.values]


@pytest.fixture(scope="module")
def model_file():
    return os.path.join("model_config_group", "base_model.yaml")


# Group constraints, i.e. those that can be defined on a system/subsystem scale
@pytest.mark.filterwarnings(
    "ignore:(?s).*All technologies were requested:calliope.exceptions.ModelWarning"
)
class TestBuildGroupConstraints:

    metadata_yaml = os.path.join(
        os.path.dirname(__file__),
        "common",
        "test_model",
        "model_config_group",
        "scenarios_metadata.yaml",
    )

    scenario_metadata = AttrDict.from_yaml(metadata_yaml)
    test_vars = [
        (k, v["constraint"], v["group_name"], v["loc_techs"])
        for k, v in scenario_metadata.items()
    ]

    @pytest.mark.parametrize("scenario, constraints, group_names, loc_techs", test_vars)
    def test_build_group_constraint(
        self, scenario, constraints, group_names, loc_techs
    ):

        scenario_model = build_model(
            model_file=os.path.join("model_config_group", "base_model.yaml"),
            scenario=scenario,
        )

        assert all(
            "group_names_" + i in scenario_model._model_data.dims for i in constraints
        )
        assert all(
            group_names[i]
            in scenario_model._model_data["group_names_" + constraints[i]].values
            for i in range(len(constraints))
        )
        assert all(
            "group_" + i in scenario_model._model_data.data_vars.keys()
            for i in constraints
        )
        assert all(
            len(
                set(loc_techs[i])
                - set(
                    scenario_model._model_data.get(
                        "group_constraint_loc_techs_" + group_names[i],
                        scenario_model._model_data.get(
                            "group_constraint_loc_tech_carriers_" + group_names[i],
                            set(),
                        ),
                    ).values
                )
            )
            == 0
            for i in range(len(group_names))
        )

        scenario_model.run(build_only=True)
        assert all(
            hasattr(scenario_model._backend_model, "group_" + i + "_constraint")
            for i in constraints
        )


@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings(
    "ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning"
)
class TestGroupConstraints:
    def test_no_group_constraint(self):
        model = build_model(model_file="group_constraints.yaml")
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply"}]
            .sum()
            .item()
        )
        assert expensive_generation == 0

    def test_switched_off_group_constraint(self):
        model = build_model(
            model_file="group_constraints.yaml",
            scenario="switching_off_group_constraint",
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply"}]
            .sum()
            .item()
        )
        assert expensive_generation == 0

    @pytest.mark.xfail(reason="Check not yet implemented.")
    def test_group_constraint_without_technology(self):
        model = build_model(
            model_file="group_constraints.yaml",
            scenario="group_constraint_without_tech",
        )
        with pytest.raises(calliope.exceptions.ModelError):
            model.run()

    def test_group_constraint_with_several_constraints(self):
        model = build_model(
            model_file="group_constraints.yaml", scenario="several_group_constraints"
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .to_dataframe()
            .reset_index()
            .groupby("techs")
            .carrier_prod.sum()
            .transform(lambda x: x / x.sum())
            .loc["expensive_elec_supply"]
        )
        assert round(expensive_generation, 5) >= 0.8


@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings(
    "ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning"
)
class TestDemandShareGroupConstraints:
    def test_no_demand_share_constraint(self):
        model = build_model(model_file="demand_share.yaml")
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .to_dataframe()
            .reset_index()
            .groupby("techs")
            .carrier_prod.sum()
            .loc["expensive_elec_supply"]
        )
        assert expensive_generation == 0

    def test_systemwide_demand_share_max_constraint(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_max_systemwide"
        )
        model.run()
        cheap_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "cheap_elec_supply", "carriers": "electricity"}]
            .sum()
        )
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (cheap_generation / demand).round(5) <= 0.3

    def test_systemwide_demand_share_min_constraint(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_min_systemwide"
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply", "carriers": "electricity"}]
            .sum()
        )
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (expensive_generation / demand).round(5) >= 0.6

    def test_systemwide_demand_share_equals_constraint(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_equals_systemwide"
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply", "carriers": "electricity"}]
            .sum()
        )
        demand = -1 * model.get_formatted_array("carrier_con").sum()
        # assert share in each timestep is 0.6
        assert (expensive_generation / demand).round(5) == 0.6

    def test_location_specific_demand_share_max_constraint(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_max_location_0"
        )
        model.run()
        generation = (
            model.get_formatted_array("carrier_prod")
            .sum(dim="timesteps")
            .loc[{"carriers": "electricity"}]
        )
        demand0 = (
            -model.get_formatted_array("carrier_con").loc[{"locs": "0"}].sum().item()
        )
        cheap_generation0 = generation.loc[
            {"locs": "0", "techs": "cheap_elec_supply"}
        ].item()
        expensive_generation1 = generation.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert round(cheap_generation0 / demand0, 5) <= 0.3
        assert expensive_generation1 == 0

    def test_location_specific_demand_share_max_constraint_two_techs(self):
        model = build_model(
            model_file="demand_share.yaml",
            scenario="demand_share_max_location_0_two_techs",
        )
        model.run()
        generation = (
            model.get_formatted_array("carrier_prod")
            .sum(dim="timesteps")
            .loc[{"carriers": "electricity"}]
        )
        demand0 = (
            -model.get_formatted_array("carrier_con").loc[{"locs": "0"}].sum().item()
        )
        generation0 = (
            generation.loc[
                {"locs": "0", "techs": ["cheap_elec_supply", "normal_elec_supply"]}
            ]
            .sum("techs")
            .item()
        )
        assert round(generation0 / demand0, 5) <= 0.4

    def test_location_specific_demand_share_min_constraint(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_min_location_0"
        )
        model.run()
        generation = (
            model.get_formatted_array("carrier_prod")
            .sum(dim="timesteps")
            .loc[{"carriers": "electricity"}]
        )
        demand0 = (
            -model.get_formatted_array("carrier_con").loc[{"locs": "0"}].sum().item()
        )
        expensive_generation0 = generation.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_generation1 = generation.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert round(expensive_generation0 / demand0, 5) >= 0.6
        assert expensive_generation1 == 0

    def test_multiple_group_constraints(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="multiple_constraints"
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(
            dim=("timesteps", "locs", "carriers")
        )
        demand = -model.get_formatted_array("carrier_con").sum().item()
        cheap_generation = generation.loc[{"techs": "cheap_elec_supply"}].item()
        expensive_generation = generation.loc[{"techs": "expensive_elec_supply"}].item()

        assert round(expensive_generation / demand, 5) >= 0.6
        assert round(cheap_generation / demand, 5) <= 0.3

    def test_multiple_group_carriers(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="multiple_carriers_max"
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(
            dim=("timesteps", "locs")
        )
        demand = -model.get_formatted_array("carrier_con").sum(
            dim=("timesteps", "locs")
        )
        cheap_generation_elec = generation.loc[
            {"techs": "cheap_elec_supply", "carriers": "electricity"}
        ].item()
        demand_elec = demand.loc[
            {"techs": "electricity_demand", "carriers": "electricity"}
        ].item()
        cheap_generation_heat = generation.loc[
            {"techs": "cheap_heat_supply", "carriers": "heat"}
        ].item()
        demand_heat = demand.loc[{"techs": "heat_demand", "carriers": "heat"}].item()

        assert round(cheap_generation_elec / demand_elec, 5) <= 0.3
        assert round(cheap_generation_heat / demand_heat, 5) <= 0.5

    def test_multiple_group_carriers_constraints(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="multiple_constraints_carriers"
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(
            dim=("timesteps", "locs")
        )
        demand = -model.get_formatted_array("carrier_con").sum(
            dim=("timesteps", "locs")
        )
        cheap_generation_elec = generation.loc[
            {"techs": "cheap_elec_supply", "carriers": "electricity"}
        ].item()
        expensive_generation_elec = generation.loc[
            {"techs": "expensive_elec_supply", "carriers": "electricity"}
        ].item()
        demand_elec = demand.loc[
            {"techs": "electricity_demand", "carriers": "electricity"}
        ].item()
        cheap_generation_heat = generation.loc[
            {"techs": "cheap_heat_supply", "carriers": "heat"}
        ].item()
        expensive_generation_heat = generation.loc[
            {"techs": "expensive_heat_supply", "carriers": "heat"}
        ].item()
        demand_heat = demand.loc[{"techs": "heat_demand", "carriers": "heat"}].item()

        assert round(cheap_generation_elec / demand_elec, 5) <= 0.3
        assert round(expensive_generation_elec / demand_elec, 5) >= 0.6
        assert round(cheap_generation_heat / demand_heat, 5) <= 0.5
        assert round(expensive_generation_heat / demand_heat, 5) >= 0.4

    def test_different_locations_per_group_constraint(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="different_locations_per_group"
        )
        model.run()
        generation = model.get_formatted_array("carrier_prod").sum(
            dim=("timesteps", "carriers")
        )
        demand = -model.get_formatted_array("carrier_con").sum(dim=("timesteps"))
        cheap_generation_0 = generation.loc[
            {"techs": "cheap_elec_supply", "locs": "0"}
        ].item()
        expensive_generation_0 = generation.loc[
            {"techs": "expensive_elec_supply", "locs": "0"}
        ].item()
        cheap_generation_1 = generation.loc[
            {"techs": "cheap_elec_supply", "locs": "1"}
        ].item()
        expensive_generation_1 = generation.loc[
            {"techs": "expensive_elec_supply", "locs": "1"}
        ].item()
        demand_elec_0 = demand.loc[
            {"techs": "electricity_demand", "carriers": "electricity", "locs": "0"}
        ].item()
        demand_elec_1 = demand.loc[
            {"techs": "electricity_demand", "carriers": "electricity", "locs": "1"}
        ].item()

        assert round(expensive_generation_0 / demand_elec_0, 5) >= 0.6
        assert expensive_generation_1 / demand_elec_1 == 0
        assert (
            round(
                (cheap_generation_0 + cheap_generation_1)
                / (demand_elec_0 + demand_elec_1),
                5,
            )
            <= 0.3
        )

    def test_transmission_not_included_in_demand(self):
        model = build_model(
            model_file="demand_share.yaml",
            scenario="transmission_not_included_in_demand",
        )
        model.run()
        assert model.results.termination_condition == "optimal"
        generation = model.get_formatted_array("carrier_prod").sum(
            dim=("timesteps", "carriers")
        )
        demand = -model.get_formatted_array("carrier_con").sum(
            dim=("timesteps", "carriers")
        )

        assert (
            generation.sel(
                locs="1",
                techs=[
                    "normal_elec_supply",
                    "cheap_elec_supply",
                    "expensive_elec_supply",
                ],
            )
            .sum(dim="techs")
            .item()
        ) == approx(0)

        cheap_elec_supply_0 = generation.sel(locs="0", techs="cheap_elec_supply").item()
        demand_0 = demand.sel(locs="0", techs="electricity_demand").item()

        assert round(cheap_elec_supply_0 / demand_0, 5) <= 0.4

    def test_demand_share_per_timestep_max(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_per_timestep_max"
        )
        model.run()
        cheap_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "cheap_elec_supply", "carriers": "electricity"}]
            .sum("locs")
        )
        demand = -1 * model.get_formatted_array("carrier_con").sum("locs")
        # assert share in each timestep is 0.6
        assert ((cheap_generation / demand).round(5) <= 0.3).all()

    def test_demand_share_per_timestep_min(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_per_timestep_min"
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply", "carriers": "electricity"}]
            .sum("locs")
        )
        demand = -1 * model.get_formatted_array("carrier_con").sum("locs")
        # assert share in each timestep is 0.6
        assert ((expensive_generation / demand).round(5) >= 0.6).all()

    def test_demand_share_per_timestep_equals(self):
        model = build_model(
            model_file="demand_share.yaml", scenario="demand_share_per_timestep_equals"
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply", "carriers": "electricity"}]
            .sum("locs")
        )
        demand = -1 * model.get_formatted_array("carrier_con").sum("locs")
        # assert share in each timestep is 0.6
        assert ((expensive_generation / demand).round(5) == 0.6).all()

    def test_demand_share_per_timestep_decision_inf(self):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_inf",
        )
        model.run()
        demand = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .sum("techs")
            .to_pandas()
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .to_pandas()
            .T
        )
        shares = supply.div(demand, axis=0)

        assert (shares["cheap_elec_supply"] == 0.1875).all()
        assert (shares["normal_elec_supply"] == 0.8125).all()

    def test_demand_share_per_timestep_decision_simple(self):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_simple",
        )
        model.run()
        demand = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .sum("techs")
            .to_pandas()
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .to_pandas()
            .T
        )
        shares = supply.div(demand, axis=0)

        assert (shares["cheap_elec_supply"] == 0.1875).all()
        assert (shares["normal_elec_supply"] == 0.8125).all()

    def test_demand_share_per_timestep_decision_not_one(self):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_not_one",
        )
        model.run()
        demand = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .sum("techs")
            .to_pandas()
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .to_pandas()
            .T
        )
        shares = supply.div(demand, axis=0)

        assert (shares["cheap_elec_supply"] == 0.1).all()
        assert (shares["normal_elec_supply"] == 0.9).all()
        assert (
            shares["expensive_elec_supply"].mean() + shares["normal_elec_supply"].mean()
            == 0.9
        )

    def test_demand_share_per_timestep_decision_per_location(self):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_per_location",
        )
        model.run()
        demand = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "electricity"}]
            .sum("techs")
            .to_series()
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .to_series()
        )
        shares = supply.div(demand, axis=0)

        assert (
            shares.loc[pd.IndexSlice["0", :, :]].unstack()["cheap_elec_supply"] == 0.25
        ).all()
        assert (
            shares.loc[pd.IndexSlice["0", :, :]].unstack()["normal_elec_supply"] == 0.75
        ).all()

        assert (
            shares.loc[pd.IndexSlice["1", :, :]].unstack()["cheap_elec_supply"] == 0.125
        ).all()
        assert (
            shares.loc[pd.IndexSlice["1", :, :]].unstack()["normal_elec_supply"]
            == 0.875
        ).all()

    def test_demand_share_per_timestep_decision_inf_with_transmission(self):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_inf,with_electricity_transmission",
        )
        model.run()
        demand = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "electricity"}]
            .sum("techs")
            .to_series()
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .to_series()
        )
        shares = supply.div(demand, axis=0)

        assert all(
            [
                i == approx(0.12373737)
                for i in shares.loc[pd.IndexSlice["0", :, :]]
                .unstack()["cheap_elec_supply"]
                .values
            ]
        )
        assert all(
            [
                i == approx(0.87626263)
                for i in shares.loc[pd.IndexSlice["0", :, :]]
                .unstack()["normal_elec_supply"]
                .values
            ]
        )
        assert (
            shares.loc[pd.IndexSlice["1", :, :]].unstack()["electricity_transmission:0"]
            == 1.0
        ).all()

    def test_demand_share_per_timestep_decision_inf_with_heat_cosntrain_electricity(
        self,
    ):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_inf,with_electricity_conversion_tech",
        )
        model.run()
        demand = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "electricity"}]
            .sum(["locs", "techs"])
            .to_pandas()
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .to_pandas()
            .T
        )
        shares = supply.div(demand, axis=0)

        assert all([i == approx(0.175926) for i in shares["cheap_elec_supply"].values])
        assert all([i == approx(0.824074) for i in shares["normal_elec_supply"].values])

    def test_demand_share_per_timestep_decision_inf_with_heat_constrain_heat(self):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_inf_with_heat,with_electricity_conversion_tech",
        )
        model.run()
        demand = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "heat"}]
            .sum(["locs", "techs"])
            .to_pandas()
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "heat"}]
            .sum("locs")
            .to_pandas()
            .T
        )
        shares = supply.div(demand, axis=0)

        assert all([i == approx(0.5) for i in shares["elec_to_heat"].values])
        assert all([i == approx(0.5) for i in shares["heating"].values])

    def test_demand_share_per_timestep_decision_inf_with_heat_constrain_heat_and_electricity(
        self,
    ):
        model = build_model(
            model_file="demand_share_decision.yaml",
            scenario="demand_share_per_timestep_decision_inf_with_heat,demand_share_per_timestep_decision_not_one,with_electricity_conversion_tech",
        )
        model.run()
        demand_heat = (
            -1
            * model.get_formatted_array("carrier_con")
            .loc[{"carriers": "heat"}]
            .sum(["locs", "techs"])
            .to_pandas()
        )
        supply_heat = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "heat"}]
            .sum("locs")
            .to_pandas()
            .T
        )
        shares_heat = supply_heat.div(demand_heat, axis=0)

        demand_elec = -1 * (
            model._model_data.carrier_con.loc[
                {
                    "loc_tech_carriers_con": [
                        i
                        for i in model._model_data.loc_tech_carriers_demand.values
                        if "electricity" in i
                    ]
                }
            ]
            .sum(["loc_tech_carriers_con"])
            .to_pandas()
        )
        supply_elec = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .to_pandas()
            .T
        )
        shares_elec = supply_elec.div(demand_elec, axis=0)

        assert all([i == approx(0.5) for i in shares_heat["elec_to_heat"].values])
        assert all([i == approx(0.5) for i in shares_heat["heating"].values])
        assert all([i == approx(0.9) for i in shares_elec["normal_elec_supply"].values])
        assert all(
            [round(i, 5) >= 0.1 for i in shares_elec["cheap_elec_supply"].values]
        )


@pytest.mark.filterwarnings(
    "ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning"
)
class TestResourceAreaGroupConstraints:
    def test_no_resource_area_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert expensive_capacity == 0

    def test_resource_area_max_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="resource_area_max_supply")
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity = capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        assert round(cheap_capacity, 5) <= 28

    def test_resource_area_min_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="resource_area_min_supply")
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert round(expensive_capacity, 5) >= 12

    def test_resource_area_equals_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="resource_area_equals_supply"
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert expensive_capacity == approx(20)

    def test_resource_area_min_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="resource_area_min_max_supply"
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity = capacity.loc[{"techs": "cheap_elec_supply"}].sum().item()
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert round(cheap_capacity, 5) <= 8
        assert round(expensive_capacity, 5) >= 12

    def test_resource_area_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="resource_area_max_supply_loc_1"
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity1 = capacity.loc[
            {"locs": "1", "techs": "cheap_elec_supply"}
        ].item()
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        assert round(cheap_capacity1, 5) <= 8
        assert expensive_capacity0 == 0

    def test_resource_area_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="resource_area_min_supply_loc_0"
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_capacity1 = capacity.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert round(expensive_capacity0, 5) >= 12
        assert expensive_capacity1 == 0

    def test_resource_area_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="resource_area_min_max_supply_loc_0_1"
        )
        model.run()
        capacity = model.get_formatted_array("resource_area")
        cheap_capacity1 = capacity.loc[
            {"locs": "1", "techs": "cheap_elec_supply"}
        ].item()
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_capacity1 = capacity.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert round(cheap_capacity1, 5) <= 8
        assert round(expensive_capacity0, 5) >= 12
        assert expensive_capacity1 == 0

    # All technologies, but insufficient resource_area_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings(
        "ignore:(?s).*['conversion', 'conversion_plus', 'demand', 'storage', "
        "'transmission'].*:calliope.exceptions.ModelWarning"
    )
    def test_resource_area_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="resource_area_max_all_techs_infeasible"
        )

        model.run()

        assert model._model_data.attrs["termination_condition"] != "optimal"


@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings(
    "ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning"
)
class TestCostCapGroupConstraint:
    def test_systemwide_cost_max_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="cheap_cost_max_systemwide"
        )
        model.run()
        cheap_cost = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "monetary", "techs": "cheap_polluting_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert round(cheap_cost, 5) <= 30

    def test_systemwide_cost_investment_max_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml",
            scenario="cheap_cost_investment_max_systemwide",
        )
        model.run()
        cheap_cost = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "monetary", "techs": "cheap_polluting_supply"}
                ]
            )
            .sum()
            .item()
        )
        cheap_cost_investment = (
            (
                model.get_formatted_array("cost_investment").loc[
                    {"costs": "monetary", "techs": "cheap_polluting_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert cheap_cost > cheap_cost_investment
        assert round(cheap_cost_investment, 5) <= 4

    def test_systemwide_cost_var_max_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="cheap_cost_var_max_systemwide"
        )
        model.run()
        cheap_cost = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "monetary", "techs": "cheap_polluting_supply"}
                ]
            )
            .sum()
            .item()
        )
        cheap_cost_var = (
            (
                model.get_formatted_array("cost_var").loc[
                    {"costs": "monetary", "techs": "cheap_polluting_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert cheap_cost > cheap_cost_var
        assert round(cheap_cost_var, 5) <= 200

    def test_systemwide_cost_min_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="expensive_cost_min_systemwide"
        )
        model.run()
        expensive_cost = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "monetary", "techs": "expensive_clean_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert round(expensive_cost, 5) >= 600

    def test_systemwide_cost_equals_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="cheap_cost_equals_systemwide"
        )
        model.run()
        cheap_cost = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "monetary", "techs": "cheap_polluting_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert cheap_cost == approx(210)

    def test_location_specific_cost_max_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="cheap_cost_max_location_0"
        )
        model.run()
        cheap_cost0 = (
            (
                model.get_formatted_array("cost").loc[
                    {
                        "costs": "monetary",
                        "techs": "cheap_polluting_supply",
                        "locs": "0",
                    }
                ]
            )
            .sum()
            .item()
        )
        assert round(cheap_cost0, 5) <= 10

    def test_systemwide_emissions_max_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="emissions_max_systemwide"
        )
        model.run()
        emissions = (
            (model.get_formatted_array("cost").loc[{"costs": "emissions"}]).sum().item()
        )
        assert round(emissions, 5) <= 400

    def test_location_specific_emissions_max_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="emissions_max_location_0"
        )
        model.run()
        emissions0 = (
            (model.get_formatted_array("cost").loc[{"costs": "emissions", "locs": "0"}])
            .sum()
            .item()
        )
        assert round(emissions0, 5) <= 200

    def test_systemwide_clean_emissions_max_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="clean_emissions_max_systemwide"
        )
        model.run()
        clean_emissions = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "emissions", "techs": "expensive_clean_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert round(clean_emissions, 5) <= 300

    def test_multiple_costs_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="multiple_costs_constraint"
        )
        model.run()
        emissions = (
            (model.get_formatted_array("cost").loc[{"costs": "emissions"}]).sum().item()
        )
        expensive_cost = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "monetary", "techs": "expensive_clean_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert round(emissions, 5) <= 400
        assert round(expensive_cost, 5) <= 600

    def test_different_locations_per_cost_group_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="different_locations_per_group"
        )
        model.run()
        cheap_cost = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "monetary", "techs": "cheap_polluting_supply"}
                ]
            )
            .sum()
            .item()
        )
        cheap_cost0 = (
            (
                model.get_formatted_array("cost").loc[
                    {
                        "costs": "monetary",
                        "techs": "cheap_polluting_supply",
                        "locs": "0",
                    }
                ]
            )
            .sum()
            .item()
        )
        assert round(cheap_cost, 5) <= 30
        assert round(cheap_cost0, 5) <= 10

    def test_different_techs_per_cost_group_constraint(self):
        model = build_model(
            model_file="model_cost_cap.yaml", scenario="different_techs_per_group"
        )
        model.run()
        emissions = (
            (model.get_formatted_array("cost").loc[{"costs": "emissions"}]).sum().item()
        )
        clean_emissions = (
            (
                model.get_formatted_array("cost").loc[
                    {"costs": "emissions", "techs": "expensive_clean_supply"}
                ]
            )
            .sum()
            .item()
        )
        assert round(emissions, 5) <= 400
        assert round(clean_emissions, 5) <= 300


@pytest.mark.xfail(reason="Tests not yet implemented.")
@pytest.mark.filterwarnings(
    "ignore:(?s).*Not all requested techs:calliope.exceptions.ModelWarning"
)
class TestSupplyShareGroupConstraints:
    def test_no_carrier_prod_share_constraint(self):
        model = build_model(model_file="carrier_prod_share.yaml")
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply"}]
            .sum()
        ).item()
        assert expensive_generation == 0

    def test_systemwide_carrier_prod_share_max_constraint(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_max_systemwide",
        )
        model.run()
        cheap_generation = (
            model.get_formatted_array("carrier_prod")
            .to_dataframe()
            .reset_index()
            .groupby("techs")
            .carrier_prod.sum()
            .transform(lambda x: x / x.sum())
            .loc["cheap_elec_supply"]
        )
        assert round(cheap_generation, 5) <= 0.4

    def test_systemwide_carrier_prod_share_min_constraint(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_min_systemwide",
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .to_dataframe()
            .reset_index()
            .groupby("techs")
            .carrier_prod.sum()
            .transform(lambda x: x / x.sum())
            .loc["expensive_elec_supply"]
        )
        assert round(expensive_generation, 5) >= 0.6

    def test_location_specific_carrier_prod_share_max_constraint(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_max_location_0",
        )
        model.run()
        generation = (
            model.get_formatted_array("carrier_prod")
            .sum(dim="timesteps")
            .loc[{"carriers": "electricity"}]
        )
        cheap_generation0 = generation.loc[
            {"locs": "0", "techs": "cheap_elec_supply"}
        ].item()
        expensive_generation0 = generation.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_generation1 = generation.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert (
            round(cheap_generation0 / (cheap_generation0 + expensive_generation0), 5)
            <= 0.4
        )
        assert expensive_generation1 == 0

    def test_location_specific_carrier_prod_share_min_constraint(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_min_location_0",
        )
        model.run()
        generation = (
            model.get_formatted_array("carrier_prod")
            .sum(dim="timesteps")
            .loc[{"carriers": "electricity"}]
        )
        cheap_generation0 = generation.loc[
            {"locs": "0", "techs": "cheap_elec_supply"}
        ].item()
        expensive_generation0 = generation.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_generation1 = generation.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert (
            round(
                expensive_generation0 / (cheap_generation0 + expensive_generation0), 5
            )
            >= 0.6
        )
        assert expensive_generation1 == 0

    def test_carrier_prod_share_with_transmission(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_min_systemwide,transmission_link",
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply"}]
            .sum()
            .item()
        )
        supply = (
            model._model_data.carrier_prod.loc[
                {
                    "loc_tech_carriers_prod": model._model_data.loc_tech_carriers_supply_conversion_all.values
                }
            ]
            .sum()
            .item()
        )

        assert round(expensive_generation / supply, 5) >= 0.6

    def test_carrier_prod_share_with_storage(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_min_systemwide,storage_tech",
        )
        model.run()
        expensive_generation = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply"}]
            .sum()
            .item()
        )
        supply = (
            model._model_data.carrier_prod.loc[
                {
                    "loc_tech_carriers_prod": model._model_data.loc_tech_carriers_supply_conversion_all.values
                }
            ]
            .sum()
            .item()
        )

        assert round(expensive_generation / supply, 5) >= 0.6

    def test_carrier_prod_share_per_timestep_max(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_per_timestep_max",
        )
        model.run()
        cheap_elec_supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "cheap_elec_supply", "carriers": "electricity"}]
            .sum("locs")
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .sum("techs")
        )
        # assert share in each timestep is 0.4
        assert ((cheap_elec_supply / supply).round(5) <= 0.4).all()

    def test_carrier_prod_share_per_timestep_min(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_per_timestep_min",
        )
        model.run()
        expensive_elec_supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply", "carriers": "electricity"}]
            .sum("locs")
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .sum("techs")
        )
        # assert share in each timestep is 0.6
        assert ((expensive_elec_supply / supply).round(5) >= 0.6).all()

    def test_carrier_prod_share_per_timestep_equals(self):
        model = build_model(
            model_file="carrier_prod_share.yaml",
            scenario="carrier_prod_share_per_timestep_equals",
        )
        model.run()
        expensive_elec_supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"techs": "expensive_elec_supply", "carriers": "electricity"}]
            .sum("locs")
        )
        supply = (
            model.get_formatted_array("carrier_prod")
            .loc[{"carriers": "electricity"}]
            .sum("locs")
            .sum("techs")
        )
        # assert share in each timestep is 0.6
        assert ((expensive_elec_supply / supply).round(5) == 0.6).all()


class TestEnergyCapShareGroupConstraints:
    @pytest.fixture(scope="module")
    def model_file(self):
        return os.path.join("model_config_group", "base_model.yaml")

    def test_no_energy_cap_share_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert expensive_capacity == 0

    def test_energy_cap_share_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_max_supply"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = capacity.loc[
            {"techs": get_supply_conversion_techs(model)}
        ]
        cheap_capacity = capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        assert (
            round(cheap_capacity / capacity_supply_conversion_all.sum().item(), 5)
            <= 0.5
        )

    def test_energy_cap_share_min_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_min_supply"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = capacity.loc[
            {"techs": get_supply_conversion_techs(model)}
        ]
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert (
            round(expensive_capacity / capacity_supply_conversion_all.sum().item(), 5)
            >= 0.4
        )

    def test_energy_cap_share_equals_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_equals_supply"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = capacity.loc[
            {"techs": get_supply_conversion_techs(model)}
        ]
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert (
            expensive_capacity / capacity_supply_conversion_all.sum().item()
            == approx(0.3)
        )

    def test_energy_cap_share_min_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_min_max_supply"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all = capacity.loc[
            {"techs": get_supply_conversion_techs(model)}
        ]
        cheap_capacity = capacity.loc[{"techs": "cheap_elec_supply"}].sum().item()
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert (
            round(cheap_capacity / capacity_supply_conversion_all.sum().item(), 5)
            <= 0.2
        )
        assert (
            round(expensive_capacity / capacity_supply_conversion_all.sum().item(), 5)
            >= 0.4
        )

    def test_energy_cap_share_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_max_supply_loc_1"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc1 = capacity.loc[
            {"locs": "1", "techs": get_supply_conversion_techs(model)}
        ]
        cheap_capacity1 = capacity.loc[
            {"locs": "1", "techs": "cheap_elec_supply"}
        ].item()
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        assert (
            round(cheap_capacity1 / capacity_supply_conversion_all_loc1.sum().item(), 5)
            <= 0.2
        )
        assert expensive_capacity0 == 0

    def test_energy_cap_share_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_min_supply_loc_0"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc0 = capacity.loc[
            {"locs": "0", "techs": get_supply_conversion_techs(model)}
        ]
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_capacity1 = capacity.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert (
            round(
                expensive_capacity0 / capacity_supply_conversion_all_loc0.sum().item(),
                5,
            )
            >= 0.4
        )
        assert expensive_capacity1 == 0

    def test_energy_cap_share_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_min_max_supply_loc_0_1"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc0 = capacity.loc[
            {"locs": "0", "techs": get_supply_conversion_techs(model)}
        ]
        capacity_supply_conversion_all_loc1 = capacity.loc[
            {"locs": "1", "techs": get_supply_conversion_techs(model)}
        ]
        cheap_capacity1 = capacity.loc[
            {"locs": "1", "techs": "cheap_elec_supply"}
        ].item()
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_capacity1 = capacity.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert (
            round(cheap_capacity1 / capacity_supply_conversion_all_loc1.sum().item(), 5)
            <= 0.2
        )
        assert (
            round(
                expensive_capacity0 / capacity_supply_conversion_all_loc0.sum().item(),
                5,
            )
            >= 0.4
        )
        assert expensive_capacity1 == 0

    # All conversion technologies with insufficient energy_cap_max to use the
    # cheap direct heat/cooling supply techs
    def test_energy_cap_share_max_non_conversion_all_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_max_non_conversion_all"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        capacity_supply_conversion_all_loc1 = capacity.loc[
            {"locs": "1", "techs": get_supply_conversion_techs(model)}
        ]
        cheap_heat = capacity.loc[{"locs": "1", "techs": "cheap_heat_supply"}].item()
        cheap_cool = capacity.loc[{"locs": "1", "techs": "cheap_cool_supply"}].item()
        assert (
            cheap_heat + cheap_cool
        ) / capacity_supply_conversion_all_loc1.sum().item() <= 0.1

    # All technologies, but insufficient energy_cap_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings(
        "ignore:(?s).*`['demand']` have been ignored*:calliope.exceptions.ModelWarning"
    )
    def test_energy_cap_share_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_share_max_all_techs_infeasible"
        )

        model.run()

        assert model._model_data.attrs["termination_condition"] != "optimal"


class TestEnergyCapGroupConstraints:
    def test_no_energy_cap_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert expensive_capacity == 0

    def test_energy_cap_max_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="energy_cap_max_supply")
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity = capacity.loc[{"techs": "cheap_elec_supply"}].sum().item()
        assert round(cheap_capacity, 5) <= 14

    def test_energy_cap_min_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="energy_cap_min_supply")
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert round(expensive_capacity, 5) >= 6

    def test_energy_cap_equals_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="energy_cap_equals_supply")
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert expensive_capacity == approx(10)

    def test_energy_cap_min_max_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="energy_cap_min_max_supply")
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity = capacity.loc[{"techs": "cheap_elec_supply"}].sum().item()
        expensive_capacity = (
            capacity.loc[{"techs": "expensive_elec_supply"}].sum().item()
        )
        assert round(cheap_capacity, 5) <= 4
        assert round(expensive_capacity, 5) >= 6

    def test_energy_cap_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_max_supply_loc_1"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity1 = capacity.loc[
            {"locs": "1", "techs": "cheap_elec_supply"}
        ].item()
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        assert round(cheap_capacity1, 5) <= 4
        assert expensive_capacity0 == 0

    def test_energy_cap_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_min_supply_loc_0"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_capacity1 = capacity.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert round(expensive_capacity0, 5) >= 6
        assert expensive_capacity1 == 0

    def test_energy_cap_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_min_max_supply_loc_0_1"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        cheap_capacity1 = capacity.loc[
            {"locs": "1", "techs": "cheap_elec_supply"}
        ].item()
        expensive_capacity0 = capacity.loc[
            {"locs": "0", "techs": "expensive_elec_supply"}
        ].item()
        expensive_capacity1 = capacity.loc[
            {"locs": "1", "techs": "expensive_elec_supply"}
        ].item()
        assert round(cheap_capacity1, 5) <= 4
        assert round(expensive_capacity0, 5) >= 6
        assert expensive_capacity1 == 0

    # All conversion technologies with insufficient energy_cap_max to use the
    # cheap direct heat/cooling supply techs
    def test_energy_cap_max_non_conversion_all_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_max_non_conversion_all"
        )
        model.run()
        capacity = model.get_formatted_array("energy_cap")
        elec_to_heat = capacity.loc[{"locs": "1", "techs": "elec_to_heat"}].item()
        elec_to_heat_cool_linked = capacity.loc[
            {"locs": "1", "techs": "elec_to_heat_cool_linked"}
        ].item()
        elec_to_heat_cool_unlinked = capacity.loc[
            {"locs": "1", "techs": "elec_to_heat_cool_unlinked"}
        ].item()
        cheap_heat = capacity.loc[{"locs": "1", "techs": "cheap_heat_supply"}].item()
        cheap_cool = capacity.loc[{"locs": "1", "techs": "cheap_cool_supply"}].item()
        assert cheap_heat == approx(0.8333, rel=0.001)
        assert cheap_cool == approx(0.1667, rel=0.001)
        assert elec_to_heat_cool_linked == approx(1.6667, rel=0.001)
        assert elec_to_heat_cool_unlinked == approx(0)
        assert elec_to_heat == approx(0)

    # All technologies, but insufficient energy_cap_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings(
        "ignore:(?s).*`['demand']` have been ignored*:calliope.exceptions.ModelWarning"
    )
    def test_energy_cap_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="energy_cap_max_all_techs_infeasible"
        )

        model.run()

        assert model._model_data.attrs["termination_condition"] != "optimal"


class TestNetImportShareGroupConstraints:
    @pytest.fixture
    def results_for_scenario(self, model_file):
        def __run(scenario=None):
            model = build_model(model_file=model_file, scenario=scenario)
            model.run()
            return model

        return __run

    @staticmethod
    def retrieve_imports(results):
        carrier_prod = results.get_formatted_array("carrier_prod").sel(
            carriers="electricity"
        )
        return carrier_prod.sel(
            techs=["transmission" in str(tech) for tech in carrier_prod.techs]
        ).sum(["techs", "timesteps"])

    @staticmethod
    def retrieve_exports(results):
        carrier_prod = results.get_formatted_array("carrier_con").sel(
            carriers="electricity"
        )
        return (
            carrier_prod.sel(
                techs=["transmission" in str(tech) for tech in carrier_prod.techs]
            ).sum(["techs", "timesteps"])
        ) * (-1)

    @staticmethod
    def retrieve_net_imports(results):
        imports = TestNetImportShareGroupConstraints.retrieve_imports(results)
        exports = TestNetImportShareGroupConstraints.retrieve_exports(results)
        return imports - exports

    @staticmethod
    def retrieve_demand(results):
        return (
            results.get_formatted_array("carrier_con")
            .sel(carriers="electricity", techs="electricity_demand")
            .sum(["timesteps"])
        )

    def test_only_imports_without_constraint(self, results_for_scenario):
        results = results_for_scenario("expensive-1")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(-demand)

    def test_only_imports_with_wrong_carrier(self, results_for_scenario):
        with pytest.warns(calliope.exceptions.ModelWarning) as warn:
            results = results_for_scenario("expensive-1,other-carrier")
        assert check_error_or_warning(
            warn,
            "Constraint group `example_net_import_share_constraint` will be completely ignored",
        )
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(-demand)

    def test_no_net_imports(self, results_for_scenario):
        results = results_for_scenario("expensive-1,no-net-imports")
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(0)

    @pytest.mark.xfail(reason="One cannot define transmission techs explicitely.")
    def test_no_imports_explicit_tech(self, results_for_scenario):
        with pytest.warns(calliope.exceptions.ModelWarning) as warn:
            results = results_for_scenario("expensive-1,no-net-imports-explicit-tech")
        assert check_error_or_warning(
            warn,
            "Constraint group `example_net_import_share_constraint` will be completely ignored",
        )
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(0)

    def test_some_imports_allowed(self, results_for_scenario):
        results = results_for_scenario("expensive-1,some-imports-allowed")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports <= -0.2 * demand

    def test_some_imports_enforced(self, results_for_scenario):
        results = results_for_scenario("expensive-1,some-imports-enforced")
        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(-0.2 * demand)

    def test_some_expensive_imports_enforced(self, results_for_scenario):
        results = results_for_scenario("expensive-1,some-expensive-imports-enforced")
        demand = self.retrieve_demand(results).sel(locs="0").item()
        net_imports = self.retrieve_net_imports(results).sel(locs="0").item()
        assert net_imports >= -0.2 * demand

    def test_ignores_imports_within_group(self, results_for_scenario):
        with pytest.warns(calliope.exceptions.ModelWarning) as warn:
            results = results_for_scenario("expensive-1,ignores-imports-within-group")
        assert check_error_or_warning(
            warn,
            "Constraint group `example_net_import_share_constraint` will be completely ignored",
        )

        demand = self.retrieve_demand(results).sel(locs="1").item()
        net_imports = self.retrieve_imports(results).sel(locs="1").item()
        assert net_imports == pytest.approx(-demand)

    def test_allows_gross_imports(self, results_for_scenario):
        results = results_for_scenario("no-net-imports,alternating-costs")
        gross_imports = self.retrieve_imports(results).sel(locs="1").item()
        assert gross_imports > 0

    def test_no_net_imports_despite_gross_imports(self, results_for_scenario):
        results = results_for_scenario("no-net-imports,alternating-costs")
        net_imports = self.retrieve_net_imports(results).sel(locs="1")
        assert net_imports <= 0


class TestCarrierProdGroupConstraints:
    def test_no_carrier_prod_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod = prod.loc[{"techs": "expensive_elec_supply"}].sum().item()
        assert expensive_prod == 0

    def test_carrier_prod_max_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="carrier_prod_max_supply")
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod = prod.loc[{"techs": "cheap_elec_supply"}].sum().item()
        assert round(cheap_prod, 5) <= 40

    def test_carrier_prod_min_supply_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="carrier_prod_min_supply")
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod = prod.loc[{"techs": "expensive_elec_supply"}].sum().item()
        assert round(expensive_prod, 5) >= 10

    def test_carrier_prod_equals_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_prod_equals_supply"
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod = prod.loc[{"techs": "expensive_elec_supply"}].sum().item()
        assert expensive_prod == approx(30)

    def test_carrier_prod_min_max_supply_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_prod_min_max_supply"
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod = prod.loc[{"techs": "cheap_elec_supply"}].sum().item()
        expensive_prod = prod.loc[{"techs": "expensive_elec_supply"}].sum().item()
        assert round(cheap_prod, 5) <= 40
        assert round(expensive_prod, 5) >= 10

    def test_carrier_prod_max_supply_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_prod_max_supply_loc_1"
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod1 = prod.loc[{"locs": "1", "techs": "cheap_elec_supply"}].sum().item()
        expensive_prod0 = (
            prod.loc[{"locs": "0", "techs": "expensive_elec_supply"}].sum().item()
        )
        assert round(cheap_prod1, 5) <= 40
        assert expensive_prod0 == 0

    def test_carrier_prod_min_supply_loc_0_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_prod_min_supply_loc_0"
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        expensive_prod0 = (
            prod.loc[{"locs": "0", "techs": "expensive_elec_supply"}].sum().item()
        )
        expensive_prod1 = (
            prod.loc[{"locs": "1", "techs": "expensive_elec_supply"}].sum().item()
        )
        assert round(expensive_prod0, 5) >= 10
        assert expensive_prod1 == 0

    def test_carrier_prod_min_max_supply_loc0_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_prod_min_max_supply_loc_0_1"
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        cheap_prod1 = prod.loc[{"locs": "1", "techs": "cheap_elec_supply"}].sum().item()
        expensive_prod0 = (
            prod.loc[{"locs": "0", "techs": "expensive_elec_supply"}].sum().item()
        )
        expensive_prod1 = (
            prod.loc[{"locs": "1", "techs": "expensive_elec_supply"}].sum().item()
        )
        assert round(cheap_prod1, 5) <= 40
        assert round(expensive_prod0, 5) >= 10
        assert expensive_prod1 == 0

    # All conversion technologies with insufficient supply_max to use the
    # cheap direct heat/cooling supply techs
    def test_carrier_prod_max_non_conversion_all_constraint(self, model_file):
        with pytest.warns(calliope.exceptions.ModelWarning) as warn:
            model = build_model(
                model_file=model_file, scenario="carrier_prod_max_non_conversion_all"
            )
        assert check_error_or_warning(
            warn,
            "Constraint group `example_carrier_prod_constraint` will be completely ignored ",
        )
        model.run()
        prod = model.get_formatted_array("carrier_prod")
        elec_to_heat = prod.loc[{"locs": "1", "techs": "elec_to_heat"}].sum().item()
        elec_to_heat_cool_linked = prod.loc[
            {"locs": "1", "techs": "elec_to_heat_cool_linked"}
        ].sum("timesteps")
        elec_to_heat_cool_unlinked = (
            prod.loc[{"locs": "1", "techs": "elec_to_heat_cool_unlinked"}].sum().item()
        )
        cheap_heat = prod.loc[{"locs": "1", "techs": "cheap_heat_supply"}].sum().item()
        cheap_cool = prod.loc[{"locs": "1", "techs": "cheap_cool_supply"}].sum().item()
        assert round(cheap_heat, 5) <= 10
        assert round(cheap_cool, 5) <= 5
        assert round(elec_to_heat_cool_linked.loc[{"carriers": "cool"}].item()) >= 1
        assert round(elec_to_heat_cool_linked.loc[{"carriers": "heat"}].item()) >= 5
        assert elec_to_heat_cool_unlinked == 0
        assert elec_to_heat == 0

    # All technologies, but insufficient supply_max for enough installed capacity to meet demand
    @pytest.mark.filterwarnings(
        "ignore:(?s).*`['demand', 'storage', 'transmission']` have been ignored*:calliope.exceptions.ModelWarning"
    )
    def test_carrier_prod_max_all_techs_infeasible_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_prod_max_all_techs_infeasible"
        )

        model.run()

        assert model._model_data.attrs["termination_condition"] != "optimal"


class TestCarrierConGroupConstraints:
    def test_no_carrier_con_constraint(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        con = model.get_formatted_array("carrier_con")
        dem = con.loc[{"techs": "electricity_demand"}].sum().item()
        assert dem == -90

    def test_carrier_con_max_demand_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="carrier_con_max_demand")
        model.run()
        con = model.get_formatted_array("carrier_con")
        dem = con.loc[{"techs": "electricity_demand"}].sum().item()
        assert round(dem, 5) >= -80

    def test_carrier_con_min_demand_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="carrier_con_min_demand")
        model.run()
        con = model.get_formatted_array("carrier_con")
        dem = con.loc[{"techs": "electricity_demand"}].sum().item()
        assert round(dem, 5) <= -40

    def test_carrier_con_equals_demand_constraint(self, model_file):
        model = build_model(model_file=model_file, scenario="carrier_con_equals_demand")
        model.run()
        con = model.get_formatted_array("carrier_con")
        dem = con.loc[{"techs": "electricity_demand"}].sum().item()
        assert dem == approx(-60)

    def test_carrier_con_max_demand_loc_1_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_con_max_demand_loc_1"
        )
        model.run()
        con = model.get_formatted_array("carrier_con")
        dem1 = con.loc[{"techs": "electricity_demand", "locs": "1"}].sum().item()
        dem0 = con.loc[{"techs": "electricity_demand", "locs": "0"}].sum().item()
        assert round(dem1, 5) >= -80
        assert dem0 == 0

    def test_carrier_con_max_multi_demand_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_con_max_multi_demand"
        )
        model.run()
        con = model.get_formatted_array("carrier_con")
        dem_heat = con.loc[{"techs": "heat_demand"}].sum().item()
        dem_cool = con.loc[{"techs": "cool_demand"}].sum().item()
        assert round(dem_heat, 5) >= -10
        assert round(dem_cool, 5) >= -5

    def test_carrier_con_min_conversion_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_con_min_conversion"
        )
        model.run()
        con = model.get_formatted_array("carrier_con")
        conversion_con = con.loc[{"techs": "elec_to_heat"}].sum().item()
        assert round(conversion_con, 5) <= -1

    def test_carrier_con_min_conversion_plus_constraint(self, model_file):
        model = build_model(
            model_file=model_file, scenario="carrier_con_min_conversion_plus"
        )
        model.run()
        con = model.get_formatted_array("carrier_con")
        conversion_con = con.loc[{"techs": "elec_to_heat_cool_unlinked"}].sum().item()
        assert round(conversion_con, 5) <= 1



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

import numpy as np

from calliope.core.attrdict import AttrDict


def concat_iterable(iterable, concatenators):
    """
    Take an iterable containing iterables of strings,
    return a list of strings concatenating inner iterables with '::'.

    E.g.:
    ``
    result = concat_iterable([('x', 'y', 'z'), ('1', '2', '3')], [':', ':'])
    result == ['x:y:z', '1:2:3']

    result = concat_iterable([('x', 'y', 'z'), ('1', '2', '3')], ['::', ':'])
    result == ['x::y:z', '1::2:3']
    ``

    """
    if len(iterable) == 0:
        return []
    concats = concatenators + [""]
    string_iter_len = len(iterable[0])
    assert all(len(i) == string_iter_len for i in iterable)
    assert string_iter_len == len(concats)

    return [
        "".join([string_iter[i] + concats[i] for i in range(string_iter_len)])
        for string_iter in iterable
    ]


def constraint_exists(model_run, loc_tech, search_entry):
    """
    Check if a cost/constraint exists for a loc_tech.

    E.g.:
    ``
    constraint_exists(model_run, 'X1::chp', 'constraints.energy_eff') will return True or False
    ``
    """
    if loc_tech in model_run.sets.loc_techs_transmission:
        loc_tech_dict = split_loc_techs_transmission(loc_tech)
        search_string = "locations.{loc_from}.links.{loc_to}.techs.{tech}.{}".format(
            search_entry, **loc_tech_dict
        )
    else:
        search_string = "locations.{0}.techs.{1}.{2}".format(
            *loc_tech.split("::"), search_entry
        )
    return model_run.get_key(search_string, None)


def get_all_carriers(config, direction="both"):
    if direction == "both":
        carrier_list = ["in", "out", "in_2", "out_2", "in_3", "out_3"]
    elif direction == "in":
        carrier_list = ["in", "in_2", "in_3"]
    elif direction == "out":
        carrier_list = ["out", "out_2", "out_3"]

    carriers = flatten_list(
        [config.get_key("carrier", "")]
        + [config.get_key("carrier_{}".format(k), "") for k in carrier_list]
    )

    return set(carriers) - set([""])


def flatten_list(unflattened_list):
    """
    Take list of iterables/non-iterables and outputs a list of non-iterables.
    """
    flattened_list = []
    for item in unflattened_list:
        if hasattr(item, "__iter__") and not isinstance(item, str):
            flattened_list.extend(item)
        else:
            flattened_list.append(item)

    return flattened_list


def split_loc_techs_transmission(transmission_string):
    """
    from loc::tech:link get out a dictionary of {loc_from:loc, loc_to:link, tech:tech}
    """
    loc, tech_link = transmission_string.split("::")
    tech, link = tech_link.split(":")

    return {"loc_from": loc, "loc_to": link, "tech": tech}


def get_systemwide_constraints(tech_config):
    if "constraints" in tech_config:
        constraints = AttrDict(
            {
                k: tech_config.constraints[k]
                for k in tech_config.constraints.keys()
                if k.endswith("_systemwide")
            }
        )
    else:
        constraints = AttrDict({})

    return constraints


def vincenty(coord1, coord2):
    """
    Vincenty's inverse method formula to calculate the distance in metres
    between two points on the surface of a spheroid (WGS84).
    modified from https://github.com/maurycyp/vincenty
    """

    a = 6378137  # equitorial radius in meters
    f = 1 / 298.257223563  # flattening from sphere to oblate spheroid
    b = a * (1 - f)  # polar radius in meters

    max_iter = 200
    thresh = 1e-12

    # short-circuit coincident points
    if coord1[0] == coord2[0] and coord1[1] == coord2[1]:
        return 0

    U1 = np.arctan((1 - f) * np.tan(np.radians(coord1[0])))
    U2 = np.arctan((1 - f) * np.tan(np.radians(coord2[0])))
    L = np.radians(coord2[1] - coord1[1])
    Lambda = L

    sinU1 = np.sin(U1)
    cosU1 = np.cos(U1)
    sinU2 = np.sin(U2)
    cosU2 = np.cos(U2)

    for iteration in range(max_iter):
        sinLambda = np.sin(Lambda)
        cosLambda = np.cos(Lambda)
        sinSigma = np.sqrt(
            (cosU2 * sinLambda) ** 2 + (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) ** 2
        )
        if sinSigma == 0:
            return 0.0  # coincident points
        cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda
        sigma = np.arctan2(sinSigma, cosSigma)
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1 - sinAlpha ** 2
        try:
            cos2SigmaM = cosSigma - 2 * sinU1 * sinU2 / cosSqAlpha
        except ZeroDivisionError:
            cos2SigmaM = 0
        C = f / 16 * cosSqAlpha * (4 + f * (4 - 3 * cosSqAlpha))
        LambdaPrev = Lambda
        Lambda = L + (1 - C) * f * sinAlpha * (
            sigma
            + C * sinSigma * (cos2SigmaM + C * cosSigma * (-1 + 2 * cos2SigmaM ** 2))
        )
        if abs(Lambda - LambdaPrev) < thresh:
            break  # successful convergence
    else:
        return None  # failure to converge

    uSq = cosSqAlpha * (a ** 2 - b ** 2) / (b ** 2)
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
    deltaSigma = (
        B
        * sinSigma
        * (
            cos2SigmaM
            + B
            / 4
            * (
                cosSigma * (-1 + 2 * cos2SigmaM ** 2)
                - B
                / 6
                * cos2SigmaM
                * (-3 + 4 * sinSigma ** 2)
                * (-3 + 4 * cos2SigmaM ** 2)
            )
        )
    )
    D = b * A * (sigma - deltaSigma)

    return round(D)



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

capacity.py
~~~~~~~~~~~~~~~~~

Capacity constraints for technologies (output, resource, area, and storage).

"""

import pyomo.core as po  # pylint: disable=import-error
import numpy as np

from calliope.backend.pyomo.util import get_param, split_comma_list
from calliope import exceptions

ORDER = 10  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]

    if backend_model.__calliope_run_config["mode"] == "operate":
        return None
    else:
        if "loc_techs_storage_capacity_constraint" in sets:
            backend_model.storage_capacity_constraint = po.Constraint(
                backend_model.loc_techs_storage_capacity_constraint,
                rule=storage_capacity_constraint_rule,
            )

        if "loc_techs_energy_capacity_storage_min_constraint" in sets:
            backend_model.energy_capacity_storage_min_constraint = po.Constraint(
                backend_model.loc_techs_energy_capacity_storage_min_constraint,
                rule=energy_capacity_storage_min_constraint_rule,
            )

        if "loc_techs_energy_capacity_storage_max_constraint" in sets:
            backend_model.energy_capacity_storage_max_constraint = po.Constraint(
                backend_model.loc_techs_energy_capacity_storage_max_constraint,
                rule=energy_capacity_storage_max_constraint_rule,
            )

        if "loc_techs_energy_capacity_storage_equals_constraint" in sets:
            backend_model.energy_capacity_storage_equals_constraint = po.Constraint(
                backend_model.loc_techs_energy_capacity_storage_equals_constraint,
                rule=energy_capacity_storage_equals_constraint_rule,
            )

        if "loc_techs_energy_capacity_storage_constraint_old" in sets:
            backend_model.energy_capacity_storage_constraint_old = po.Constraint(
                backend_model.loc_techs_energy_capacity_storage_constraint_old,
                rule=energy_capacity_storage_constraint_rule_old,
            )

        if "loc_techs_resource_capacity_constraint" in sets:
            backend_model.resource_capacity_constraint = po.Constraint(
                backend_model.loc_techs_resource_capacity_constraint,
                rule=resource_capacity_constraint_rule,
            )

        if "loc_techs_resource_capacity_equals_energy_capacity_constraint" in sets:
            backend_model.resource_capacity_equals_energy_capacity_constraint = po.Constraint(
                backend_model.loc_techs_resource_capacity_equals_energy_capacity_constraint,
                rule=resource_capacity_equals_energy_capacity_constraint_rule,
            )

        if "loc_techs_resource_area_constraint" in sets:
            backend_model.resource_area_constraint = po.Constraint(
                backend_model.loc_techs_resource_area_constraint,
                rule=resource_area_constraint_rule,
            )

        if "loc_techs_resource_area_per_energy_capacity_constraint" in sets:
            backend_model.resource_area_per_energy_capacity_constraint = po.Constraint(
                backend_model.loc_techs_resource_area_per_energy_capacity_constraint,
                rule=resource_area_per_energy_capacity_constraint_rule,
            )

        if "locs_resource_area_capacity_per_loc_constraint" in sets:
            backend_model.resource_area_capacity_per_loc_constraint = po.Constraint(
                backend_model.locs_resource_area_capacity_per_loc_constraint,
                rule=resource_area_capacity_per_loc_constraint_rule,
            )

        if "loc_techs_energy_capacity_constraint" in sets:
            backend_model.energy_capacity_constraint = po.Constraint(
                backend_model.loc_techs_energy_capacity_constraint,
                rule=energy_capacity_constraint_rule,
            )

        if "techs_energy_capacity_systemwide_constraint" in sets:
            backend_model.energy_capacity_systemwide_constraint = po.Constraint(
                backend_model.techs_energy_capacity_systemwide_constraint,
                rule=energy_capacity_systemwide_constraint_rule,
            )


def get_capacity_constraint(
    backend_model, parameter, loc_tech, _equals=None, _max=None, _min=None, scale=None
):

    decision_variable = getattr(backend_model, parameter)

    if not _equals:
        _equals = get_param(backend_model, parameter + "_equals", loc_tech)
    if not _max:
        _max = get_param(backend_model, parameter + "_max", loc_tech)
    if not _min:
        _min = get_param(backend_model, parameter + "_min", loc_tech)
    if po.value(_equals) is not False and po.value(_equals) is not None:
        if np.isinf(po.value(_equals)):
            e = exceptions.ModelError
            raise e(
                "Cannot use inf for {}_equals for loc:tech `{}`".format(
                    parameter, loc_tech
                )
            )
        if scale:
            _equals *= scale
        return decision_variable[loc_tech] == _equals
    else:
        if po.value(_min) == 0 and np.isinf(po.value(_max)):
            return po.Constraint.NoConstraint
        else:
            if scale:
                _max *= scale
                _min *= scale
            return (_min, decision_variable[loc_tech], _max)


def storage_capacity_constraint_rule(backend_model, loc_tech):
    """
    Set maximum storage capacity. Supply_plus & storage techs only

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{cap}}(loc::tech)
            \\begin{cases}
                = storage_{cap, equals}(loc::tech),& \\text{if } storage_{cap, equals}(loc::tech)\\\\
                \\leq storage_{cap, max}(loc::tech),& \\text{if } storage_{cap, max}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall loc::tech \\in loc::techs_{store}

    and (if ``equals`` not enforced):

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{cap}}(loc::tech) \\geq storage_{cap, min}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{store}

    """
    return get_capacity_constraint(backend_model, "storage_cap", loc_tech)


def energy_capacity_storage_constraint_rule_old(backend_model, loc_tech):
    """
    Set an additional energy capacity constraint on storage technologies,
    based on their use of `charge_rate`.

    This is deprecated and will be removed in Calliope 0.7.0. Instead of
    `charge_rate`, please use `energy_cap_per_storage_cap_max`.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{energy_{cap}}(loc::tech)
            \\leq \\boldsymbol{storage_{cap}}(loc::tech) \\times charge\\_rate(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{store}

    """
    charge_rate = get_param(backend_model, "charge_rate", loc_tech)

    return backend_model.energy_cap[loc_tech] <= (
        backend_model.storage_cap[loc_tech] * charge_rate
    )


def energy_capacity_storage_min_constraint_rule(backend_model, loc_tech):
    """
    Limit energy capacities of storage technologies based on their storage capacities.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{energy_{cap}}(loc::tech)
                \\geq \\boldsymbol{storage_{cap}}(loc::tech) \\times energy\\_cap\\_per\\_storage\\_cap\\_min(loc::tech)\\\\
            \\forall loc::tech \\in loc::techs_{store}

    """
    return backend_model.energy_cap[loc_tech] >= (
        backend_model.storage_cap[loc_tech]
        * get_param(backend_model, "energy_cap_per_storage_cap_min", loc_tech)
    )


def energy_capacity_storage_max_constraint_rule(backend_model, loc_tech):
    """
    Limit energy capacities of storage technologies based on their storage capacities.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{energy_{cap}}(loc::tech)
                \\leq \\boldsymbol{storage_{cap}}(loc::tech) \\times energy\\_cap\\_per\\_storage\\_cap\\_max(loc::tech)\\\\
            \\forall loc::tech \\in loc::techs_{store}

    """
    return backend_model.energy_cap[loc_tech] <= (
        backend_model.storage_cap[loc_tech]
        * get_param(backend_model, "energy_cap_per_storage_cap_max", loc_tech)
    )


def energy_capacity_storage_equals_constraint_rule(backend_model, loc_tech):
    """
    Limit energy capacities of storage technologies based on their storage capacities.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{energy_{cap}}(loc::tech)
                = \\boldsymbol{storage_{cap}}(loc::tech) \\times energy\\_cap\\_per\\_storage\\_cap\\_equals(loc::tech)
            \\forall loc::tech \\in loc::techs_{store}

    """
    return backend_model.energy_cap[loc_tech] == (
        backend_model.storage_cap[loc_tech]
        * get_param(backend_model, "energy_cap_per_storage_cap_equals", loc_tech)
    )


def resource_capacity_constraint_rule(backend_model, loc_tech):
    """
    Add upper and lower bounds for resource_cap.

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{cap}}(loc::tech)
            \\begin{cases}
                = resource_{cap, equals}(loc::tech),& \\text{if } resource_{cap, equals}(loc::tech)\\\\
                \\leq resource_{cap, max}(loc::tech),& \\text{if } resource_{cap, max}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall loc::tech \\in loc::techs_{finite\\_resource\\_supply\\_plus}

    and (if ``equals`` not enforced):

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{cap}}(loc::tech) \\geq resource_{cap, min}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{finite\\_resource\\_supply\\_plus}
    """

    return get_capacity_constraint(backend_model, "resource_cap", loc_tech)


def resource_capacity_equals_energy_capacity_constraint_rule(backend_model, loc_tech):
    """
    Add equality constraint for resource_cap to equal energy_cap, for any technologies
    which have defined resource_cap_equals_energy_cap.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{cap}}(loc::tech) = \\boldsymbol{energy_{cap}}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{finite\\_resource\\_supply\\_plus}
            \\text{ if } resource\\_cap\\_equals\\_energy\\_cap = \\text{True}
    """
    return backend_model.resource_cap[loc_tech] == backend_model.energy_cap[loc_tech]


def resource_area_constraint_rule(backend_model, loc_tech):
    """
    Set upper and lower bounds for resource_area.

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{area}}(loc::tech)
            \\begin{cases}
                = resource_{area, equals}(loc::tech),& \\text{if } resource_{area, equals}(loc::tech)\\\\
                \\leq resource_{area, max}(loc::tech),& \\text{if } resource_{area, max}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall loc::tech \\in loc::techs_{area}

    and (if ``equals`` not enforced):

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{area}}(loc::tech) \\geq resource_{area, min}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{area}
    """
    energy_cap_max = get_param(backend_model, "energy_cap_max", loc_tech)
    area_per_energy_cap = get_param(
        backend_model, "resource_area_per_energy_cap", loc_tech
    )

    if po.value(energy_cap_max) == 0 and not po.value(area_per_energy_cap):
        # If a technology has no energy_cap here, we force resource_area to zero,
        # so as not to accrue spurious costs
        return backend_model.resource_area[loc_tech] == 0
    else:
        return get_capacity_constraint(backend_model, "resource_area", loc_tech)


def resource_area_per_energy_capacity_constraint_rule(backend_model, loc_tech):
    """
    Add equality constraint for resource_area to equal a percentage of energy_cap,
    for any technologies which have defined resource_area_per_energy_cap

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{area}}(loc::tech) =
            \\boldsymbol{energy_{cap}}(loc::tech) \\times area\\_per\\_energy\\_cap(loc::tech)
            \\quad \\forall loc::tech \\in locs::techs_{area} \\text{ if } area\\_per\\_energy\\_cap(loc::tech)
    """
    area_per_energy_cap = get_param(
        backend_model, "resource_area_per_energy_cap", loc_tech
    )

    return (
        backend_model.resource_area[loc_tech]
        == backend_model.energy_cap[loc_tech] * area_per_energy_cap
    )


def resource_area_capacity_per_loc_constraint_rule(backend_model, loc):
    """
    Set upper bound on use of area for all locations which have `available_area`
    constraint set. Does not consider resource_area applied to demand technologies

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{tech} \\boldsymbol{resource_{area}}(loc::tech) \\leq available\\_area
            \\quad \\forall loc \\in locs \\text{ if } available\\_area(loc)
    """
    model_data_dict = backend_model.__calliope_model_data["data"]
    available_area = model_data_dict["available_area"][loc]

    loc_techs = split_comma_list(model_data_dict["lookup_loc_techs_area"][loc])

    return (
        sum(backend_model.resource_area[loc_tech] for loc_tech in loc_techs)
        <= available_area
    )


def energy_capacity_constraint_rule(backend_model, loc_tech):
    """
    Set upper and lower bounds for energy_cap.

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\frac{\\boldsymbol{energy_{cap}}(loc::tech)}{energy_{cap, scale}(loc::tech)}
            \\begin{cases}
                = energy_{cap, equals}(loc::tech),& \\text{if } energy_{cap, equals}(loc::tech)\\\\
                \\leq energy_{cap, max}(loc::tech),& \\text{if } energy_{cap, max}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall loc::tech \\in loc::techs

    and (if ``equals`` not enforced):

    .. container:: scrolling-wrapper

        .. math::

            \\frac{\\boldsymbol{energy_{cap}}(loc::tech)}{energy_{cap, scale}(loc::tech)}
            \\geq energy_{cap, min}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs
    """
    scale = get_param(backend_model, "energy_cap_scale", loc_tech)
    return get_capacity_constraint(backend_model, "energy_cap", loc_tech, scale=scale)


def energy_capacity_systemwide_constraint_rule(backend_model, tech):
    """
    Set constraints to limit the capacity of a single technology type across all locations in the model.

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc}\\boldsymbol{energy_{cap}}(loc::tech)
            \\begin{cases}
                = energy_{cap, equals, systemwide}(loc::tech),&
                    \\text{if } energy_{cap, equals, systemwide}(loc::tech)\\\\
                \\leq energy_{cap, max, systemwide}(loc::tech),&
                    \\text{if } energy_{cap, max, systemwide}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall tech \\in techs

    """

    if tech in getattr(backend_model, "techs_transmission_names", []):
        all_loc_techs = [
            i
            for i in backend_model.loc_techs_transmission
            if i.split("::")[1].split(":")[0] == tech
        ]
        multiplier = 2  # there are always two technologies associated with one link
    else:
        all_loc_techs = [i for i in backend_model.loc_techs if i.split("::")[1] == tech]
        multiplier = 1

    max_systemwide = get_param(backend_model, "energy_cap_max_systemwide", tech)
    equals_systemwide = get_param(backend_model, "energy_cap_equals_systemwide", tech)

    if np.isinf(po.value(max_systemwide)) and not equals_systemwide:
        return po.Constraint.NoConstraint
    elif equals_systemwide and np.isinf(po.value(equals_systemwide)):
        raise exceptions.ModelError(
            "Cannot use inf for energy_cap_equals_systemwide for tech `{}`".format(tech)
        )

    sum_expr = sum(backend_model.energy_cap[loc_tech] for loc_tech in all_loc_techs)

    if equals_systemwide:
        return sum_expr == equals_systemwide * multiplier
    else:
        return sum_expr <= max_systemwide * multiplier



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

util.py
~~~~~~~

Analysis util functions.

"""


def subset_sum_squeeze(data, subset={}, sum_dims=None, squeeze=False):
    """
    Take an xarray DataArray and apply indexing, summing and squeezing,
    to prepare it for analysis.

    Parameters
    ----------
    data : xarray DataArray
        A Calliope model data variable, either input or output, which has been
        reformatted to deconcatenate loc_techs (or loc_tech_carriers/loc_carriers)
        using calliope.Model().get_formatted_array(original_data)
    subset : dict, default {}
        key:value pairs for indexing data. Uses xarray `loc[]` to index.
    sum_dims : str or list of strings, default None
        Names of dimensions over which to sum the data.
    squeeze : bool, str, or list of strings, default False
        If True, remove all dimensions of length 1
        If string, try to remove that dimension, if it is of length 1
        If list of strings, try to remove all listed dimensions, if they are length 1

    Returns
    -------
    data : xarray DataArray

    Examples
    --------
    (in) data = carrier_prod, dimensions = (locs: 2, techs: 5, carriers: 1, timesteps: 100)
    subset_sum_squeeze(
        data, subset={'techs': ['ccgt', 'csp']}, sum_dims='locs', squeeze=True
    )
    (out) data = carrier_prod, dimensions = (techs: 2, timesteps: 100)

    """
    if subset:  # first, subset the data
        allowed_subsets = {k: v for k, v in subset.items() if k in data.dims}
        allowed_items = {}

        for k, v in allowed_subsets.items():
            if isinstance(v, str):
                v = [v]
            allowed_items[k] = [i for i in v if i in data[k].values]

        data = data.loc[allowed_items]

    if sum_dims:  # second, sum along all necessary dimensions
        data = data.sum(sum_dims)

    if squeeze:  # finally, squeeze out single length dimensions
        if len(data.techs) == 1:
            dims_to_squeeze = [
                i for i in data.dims if len(data[i]) == 1 and i != "techs"
            ]
            data = data.squeeze(dims_to_squeeze)
        else:
            data = data.squeeze()

    return data



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

from copy import deepcopy
import functools
import importlib
import operator
import os
import sys


def get_from_dict(data_dict, map_list):
    return functools.reduce(operator.getitem, map_list, data_dict)


def apply_to_dict(data_dict, map_list, func, args):
    getattr(get_from_dict(data_dict, map_list[:-1])[map_list[-1]], func)(*args)


memoize = functools.lru_cache(maxsize=2048)


class memoize_instancemethod(object):
    """
    Cache the return value of a method on a per-instance basis
    (as opposed to a per-class basis like functools.lru_cache does)

    Source: http://code.activestate.com/recipes/577452/

    This class is meant to be used as a decorator of methods. The return
    value from a given method invocation will be cached on the instance
    whose method was invoked. All arguments passed to a method decorated
    with memoize must be hashable.

    If a memoized method is invoked directly on its class the result
    will not be cached. Instead the method will be invoked like a
    static method.

    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return functools.partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(list(kw.items())))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


def relative_path(base_path_file, path):
    """
    If ``path`` is not absolute, it is interpreted as relative to the
    path of the given ``base_path_file``.

    """
    # Check if base_path_file is a string because it might be an AttrDict
    if not os.path.isabs(path) and isinstance(base_path_file, str):
        path = os.path.join(os.path.dirname(base_path_file), path)
    return path


def load_function(source):
    """
    Returns a function from a module, given a source string of the form:

        'module.submodule.subsubmodule.function_name'

    """
    module_string, function_string = source.rsplit(".", 1)
    modules = sys.modules.keys()
    # Check if module already loaded, if so, don't re-import it
    if module_string in modules:
        module = sys.modules[module_string]
    # Else load the module
    else:
        module = importlib.import_module(module_string)
    return getattr(module, function_string)


def plugin_load(name, builtin_module):
    try:  # First try importing as a third-party module
        func = load_function(name)
    except ValueError:
        # ValueError raised if we got a string without '.',
        # which implies a builtin function,
        # so we attempt to load from the given module
        func_string = builtin_module + "." + name
        func = load_function(func_string)
    return func



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

costs.py
~~~~~~~~

Cost constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import (
    get_param,
    get_timestep_weight,
    loc_tech_is_in,
    invalid,
)

ORDER = 10  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]
    run_config = backend_model.__calliope_run_config

    if "loc_techs_cost_constraint" in sets:
        backend_model.cost_constraint = po.Constraint(
            backend_model.costs, backend_model.loc_techs_cost, rule=cost_constraint_rule
        )

    # FIXME: remove check for operate from constraint files, avoid investment costs more intelligently?
    if (
        "loc_techs_cost_investment_constraint" in sets
        and run_config["mode"] != "operate"
    ):
        # Right-hand side expression can be later updated by MILP investment costs
        backend_model.cost_investment_rhs = po.Expression(
            backend_model.costs,
            backend_model.loc_techs_cost_investment_constraint,
            initialize=0.0,
        )

        backend_model.cost_investment_constraint = po.Constraint(
            backend_model.costs,
            backend_model.loc_techs_cost_investment_constraint,
            rule=cost_investment_constraint_rule,
        )

    if "loc_techs_om_cost" in sets:
        # Right-hand side expression can be later updated by export costs/revenue
        backend_model.cost_var_rhs = po.Expression(
            backend_model.costs,
            backend_model.loc_techs_om_cost,
            backend_model.timesteps,
            initialize=0.0,
        )
    if "loc_techs_cost_var_constraint" in sets:
        # Constraint is built over a different loc_techs set to expression, as
        # it is updated in conversion.py and conversion_plus.py constraints
        backend_model.cost_var_constraint = po.Constraint(
            backend_model.costs,
            backend_model.loc_techs_cost_var_constraint,
            backend_model.timesteps,
            rule=cost_var_constraint_rule,
        )


def cost_constraint_rule(backend_model, cost, loc_tech):
    """
    Combine investment and time varying costs into one cost per technology.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost}(cost, loc::tech) = \\boldsymbol{cost_{investment}}(cost, loc::tech)
            + \\sum_{timestep \\in timesteps} \\boldsymbol{cost_{var}}(cost, loc::tech, timestep)

    """
    run_config = backend_model.__calliope_run_config

    # FIXME: remove check for operate from constraint files, avoid investment costs more intelligently?
    if (
        loc_tech_is_in(backend_model, loc_tech, "loc_techs_investment_cost")
        and run_config["mode"] != "operate"
    ):
        cost_investment = backend_model.cost_investment[cost, loc_tech]
    else:
        cost_investment = 0

    if loc_tech_is_in(backend_model, loc_tech, "loc_techs_om_cost"):
        cost_var = sum(
            backend_model.cost_var[cost, loc_tech, timestep]
            for timestep in backend_model.timesteps
        )
    else:
        cost_var = 0

    return backend_model.cost[cost, loc_tech] == cost_investment + cost_var


def cost_investment_constraint_rule(backend_model, cost, loc_tech):
    """
    Calculate costs from capacity decision variables.

    Transmission technologies "exist" at two locations, so their cost
    is divided by 2.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{investment}}(cost, loc::tech) =
            cost_{fractional\\_om}(cost, loc::tech) +
            cost_{fixed\\_om}(cost, loc::tech) + cost_{cap}(cost, loc::tech)

            cost_{cap}(cost, loc::tech) =
            depreciation\\_rate * ts\\_weight *
            (cost_{energy\\_cap}(cost, loc::tech) \\times \\boldsymbol{energy_{cap}}(loc::tech)
            + cost_{storage\\_cap}(cost, loc::tech) \\times \\boldsymbol{storage_{cap}}(loc::tech)
            + cost_{resource\\_cap}(cost, loc::tech) \\times \\boldsymbol{resource_{cap}}(loc::tech)
            + cost_{resource\\_area}(cost, loc::tech)) \\times \\boldsymbol{resource_{area}}(loc::tech)

            depreciation\\_rate =
            \\begin{cases}
                = 1 / plant\\_life,&
                    \\text{if } interest\\_rate = 0\\\\
                = \\frac{interest\\_rate \\times (1 + interest\\_rate)^{plant\\_life}}{(1 + interest\\_rate)^{plant\\_life} - 1},&
                    \\text{if } interest\\_rate \\gt 0\\\\
            \\end{cases}

            ts\\_weight = \\sum_{timestep \\in timesteps} (time\\_res(timestep) \\times weight(timestep)) \\times \\frac{1}{8760}

    """
    model_data_dict = backend_model.__calliope_model_data

    def _get_investment_cost(capacity_decision_variable, calliope_set):
        """
        Conditionally add investment costs, if the relevant set of technologies
        exists. Both inputs are strings.
        """
        if loc_tech_is_in(backend_model, loc_tech, calliope_set):
            _cost = getattr(backend_model, capacity_decision_variable)[
                loc_tech
            ] * get_param(
                backend_model, "cost_" + capacity_decision_variable, (cost, loc_tech)
            )
            return _cost
        else:
            return 0

    cost_energy_cap = backend_model.energy_cap[loc_tech] * get_param(
        backend_model, "cost_energy_cap", (cost, loc_tech)
    )

    cost_storage_cap = _get_investment_cost("storage_cap", "loc_techs_store")
    cost_resource_cap = _get_investment_cost("resource_cap", "loc_techs_supply_plus")
    cost_resource_area = _get_investment_cost("resource_area", "loc_techs_area")

    cost_om_annual_investment_fraction = get_param(
        backend_model, "cost_om_annual_investment_fraction", (cost, loc_tech)
    )
    cost_om_annual = get_param(backend_model, "cost_om_annual", (cost, loc_tech))

    ts_weight = get_timestep_weight(backend_model)
    depreciation_rate = model_data_dict["data"]["cost_depreciation_rate"].get(
        (cost, loc_tech), 0
    )

    cost_cap = (
        depreciation_rate
        * ts_weight
        * (cost_energy_cap + cost_storage_cap + cost_resource_cap + cost_resource_area)
    )

    # Transmission technologies exist at two locations, thus their cost is divided by 2
    if loc_tech_is_in(backend_model, loc_tech, "loc_techs_transmission"):
        cost_cap = cost_cap / 2

    cost_fractional_om = cost_om_annual_investment_fraction * cost_cap
    cost_fixed_om = cost_om_annual * backend_model.energy_cap[loc_tech] * ts_weight

    backend_model.cost_investment_rhs[cost, loc_tech].expr = (
        cost_fractional_om + cost_fixed_om + cost_cap
    )

    return (
        backend_model.cost_investment[cost, loc_tech]
        == backend_model.cost_investment_rhs[cost, loc_tech]
    )


def cost_var_constraint_rule(backend_model, cost, loc_tech, timestep):
    """
    Calculate costs from time-varying decision variables

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{var}}(cost, loc::tech, timestep) = cost_{prod}(cost, loc::tech, timestep) + cost_{con}(cost, loc::tech, timestep)

            cost_{prod}(cost, loc::tech, timestep) = cost_{om\\_prod}(cost, loc::tech, timestep) \\times weight(timestep) \\times \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)

            prod\\_con\\_eff =
            \\begin{cases}
                = \\boldsymbol{resource_{con}}(loc::tech, timestep),&
                    \\text{if } loc::tech \\in loc\\_techs\\_supply\\_plus \\\\
                = \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{energy_eff(loc::tech, timestep)},&
                    \\text{if } loc::tech \\in loc\\_techs\\_supply \\\\
            \\end{cases}

            cost_{con}(cost, loc::tech, timestep) = cost_{om\\_con}(cost, loc::tech, timestep) \\times weight(timestep) \\times prod\\_con\\_eff

    """
    model_data_dict = backend_model.__calliope_model_data

    cost_om_prod = get_param(backend_model, "cost_om_prod", (cost, loc_tech, timestep))
    cost_om_con = get_param(backend_model, "cost_om_con", (cost, loc_tech, timestep))
    weight = backend_model.timestep_weights[timestep]

    loc_tech_carrier = model_data_dict["data"]["lookup_loc_techs"][loc_tech]

    if loc_tech_is_in(
        backend_model, loc_tech_carrier, "loc_tech_carriers_prod"
    ) and not invalid(cost_om_prod):
        cost_prod = (
            cost_om_prod
            * weight
            * backend_model.carrier_prod[loc_tech_carrier, timestep]
        )
    else:
        cost_prod = 0

    cost_con = 0
    if not invalid(cost_om_con):
        if loc_tech_is_in(backend_model, loc_tech, "loc_techs_supply_plus"):
            cost_con = (
                cost_om_con * weight * backend_model.resource_con[loc_tech, timestep]
            )
        elif loc_tech_is_in(backend_model, loc_tech, "loc_techs_supply"):
            energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))
            # in case energy_eff is zero, to avoid an infinite value
            if po.value(energy_eff) > 0:
                cost_con = (
                    cost_om_con
                    * weight
                    * (
                        backend_model.carrier_prod[loc_tech_carrier, timestep]
                        / energy_eff
                    )
                )
        elif loc_tech_is_in(backend_model, loc_tech, "loc_techs_demand"):
            cost_con = (
                cost_om_con
                * weight
                * (-1)
                * backend_model.carrier_con[loc_tech_carrier, timestep]
            )

    backend_model.cost_var_rhs[cost, loc_tech, timestep].expr = cost_prod + cost_con
    return (
        backend_model.cost_var[cost, loc_tech, timestep]
        == backend_model.cost_var_rhs[cost, loc_tech, timestep]
    )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

cli.py
~~~~~~

Command-line interface.

"""

import contextlib
import datetime
import itertools
import logging
import os
import pstats
import shutil
import sys
import traceback

import click

from calliope import AttrDict, Model, examples, read_netcdf
from calliope._version import __version__
from calliope.core.util.generate_runs import generate
from calliope.core.util.logging import set_log_verbosity
from calliope.exceptions import BackendError

_time_format = "%Y-%m-%d %H:%M:%S"


_debug = click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Print debug information when encountering errors.",
)

_quiet = click.option(
    "--quiet",
    is_flag=True,
    default=False,
    help="Be less verbose about what is happening, including hiding " "solver output.",
)

_pdb = click.option(
    "--pdb",
    is_flag=True,
    default=False,
    help="If used together with --debug, drop into interactive "
    "debugger on encountering errors.",
)

_profile = click.option(
    "--profile", is_flag=True, default=False, help="Run through cProfile."
)

_profile_filename = click.option(
    "--profile_filename",
    type=str,
    help="Filename to save profile to if enabled --profile.",
)


_fail_when_infeasible = click.option(
    "--fail_when_infeasible/--no_fail_when_infeasible",
    is_flag=True,
    default=True,
    help="Return fail on command line when problem is infeasible (default True).",
)


@contextlib.contextmanager
def format_exceptions(
    debug=False, pdb=False, profile=False, profile_filename=None, start_time=None
):

    try:
        if profile:
            import cProfile

            profile = cProfile.Profile()
            profile.enable()
        yield
        if profile:
            profile.disable()
            if profile_filename:
                dump_path = os.path.expanduser(profile_filename)
                click.secho("\nSaving cProfile output to: {}".format(dump_path))
                profile.dump_stats(dump_path)
            else:
                click.secho("\n\n----PROFILE OUTPUT----\n\n")
                stats = pstats.Stats(profile).sort_stats("cumulative")
                stats.print_stats(20)  # Print first 20 lines

    except Exception as e:
        if debug:
            traceback.print_exc()
            if pdb:
                import ipdb

                ipdb.post_mortem(e.__traceback__)
        else:
            stack = traceback.extract_tb(e.__traceback__)
            # Get last stack trace entry still in Calliope
            last = [i for i in stack if "calliope" in i[0]][-1]
            err_string = "\nError in {}, {}:{}".format(last[2], last[0], last[1])
            click.secho(err_string, fg="red")
            click.secho(str(e), fg="red")
            if start_time:
                print_end_time(start_time, msg="aborted due to an error")
        sys.exit(1)


def print_end_time(start_time, msg="complete"):
    end_time = datetime.datetime.now()
    secs = round((end_time - start_time).total_seconds(), 1)
    tend = end_time.strftime(_time_format)
    click.secho(
        "\nCalliope run {}. "
        "Elapsed: {} seconds (time at exit: {})".format(msg, secs, tend)
    )


def _get_version():
    return "Version {}".format(__version__)


def _cli_start(debug, quiet):
    """
    Initial setup for CLI commands.
    Returns ``start_time`` (datetime timestamp)

    """

    if debug:
        click.secho(_get_version())

    if debug:
        verbosity = "debug"
        log_solver = True
    else:
        if quiet:
            verbosity = "warning"
            log_solver = False
        else:  # Default option
            verbosity = "info"
            log_solver = True

    set_log_verbosity(
        verbosity, include_solver_output=log_solver, capture_warnings=True
    )

    start_time = datetime.datetime.now()

    return start_time


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("--version", is_flag=True, default=False, help="Display version.")
def cli(ctx, version):
    """Calliope: a multi-scale energy systems modelling framework"""
    if ctx.invoked_subcommand is None and not version:
        click.secho(ctx.get_help())
    if version:
        click.secho(_get_version())


@cli.command(name="new", short_help="Create a new model based on a built-in example.")
@click.argument("path")
@click.option("--template", type=str, default=None)
@_debug
def new(path, template, debug):
    """
    Create new model at the given ``path``, based on one of the built-in
    example models. The target path must not yet exist. Intermediate
    directories will be created automatically.
    """
    _cli_start(debug, quiet=False)

    with format_exceptions(debug):
        if template is None:
            template = "national_scale"
        source_path = examples._PATHS[template]
        click.echo("Copying {} template to target directory: {}".format(template, path))
        shutil.copytree(source_path, path)


def _run_setup_model(model_file, scenario, model_format, override_dict):
    """
    Build model in CLI commands. Returns ``model``, a ready-to-run
    calliope.Model instance.

    """
    # Try to determine model file type if not given explicitly
    if model_format is None:
        if model_file.split(".")[-1] in ["yaml", "yml"]:
            model_format = "yaml"
        elif model_file.split(".")[-1] in ["nc", "nc4", "netcdf"]:
            model_format = "netcdf"
        else:
            raise ValueError(
                "Cannot determine model file format based on file "
                'extension for "{}". Set format explicitly with '
                "--model_format.".format(model_file)
            )

    if model_format == "yaml":
        model = Model(model_file, scenario=scenario, override_dict=override_dict)
    elif model_format == "netcdf":
        if scenario is not None or override_dict is not None:
            raise ValueError(
                "When loading a pre-built model from NetCDF, the "
                "--scenario and --override_dict options are not available."
            )
        model = read_netcdf(model_file)
    else:
        raise ValueError("Invalid model format: {}".format(model_format))

    return model


@cli.command(name="run", short_help="Build and run a model.")
@click.argument("model_file")
@click.option("--scenario")
@click.option("--model_format")
@click.option("--override_dict")
@click.option("--save_netcdf")
@click.option("--save_csv")
@click.option("--save_plots")
@click.option("--save_logs")
@click.option(
    "--save_lp",
    help="Build and save model to the given LP file. "
    "When this is set, the model is not sent to a solver, and all other save options are ignored.",
)
@_debug
@_quiet
@_pdb
@_profile
@_profile_filename
@_fail_when_infeasible
def run(
    model_file,
    scenario,
    model_format,
    override_dict,
    save_netcdf,
    save_csv,
    save_plots,
    save_logs,
    save_lp,
    debug,
    quiet,
    pdb,
    profile,
    profile_filename,
    fail_when_infeasible,
):
    """
    Execute the given model. Tries to guess from the file extension whether
    ``model_file`` is a YAML file or a pre-built model saved to NetCDF.
    This can also explicitly be set with the --model_format=yaml or
    --model_format=netcdf option.

    """
    start_time = _cli_start(debug, quiet)
    click.secho(
        "Calliope {} starting at {}\n".format(
            __version__, start_time.strftime(_time_format)
        )
    )

    with format_exceptions(debug, pdb, profile, profile_filename, start_time):

        model = _run_setup_model(model_file, scenario, model_format, override_dict)
        click.secho(model.info() + "\n")

        # Only save LP file
        if save_lp:  # Only save LP file without solving model
            click.secho("Saving model to LP file...")
            if save_csv is not None or save_netcdf is not None:
                click.secho(
                    "WARNING: Model will not be solved - ignoring other save options!",
                    fg="red",
                    bold=True,
                )
            model.to_lp(save_lp)
            print_end_time(start_time)

        # Else run the model, then save outputs
        else:
            click.secho("Starting model run...")

            if save_logs:
                model.run_config["save_logs"] = save_logs

            if save_csv is None and save_netcdf is None and save_plots is not None:
                click.secho(
                    "\n!!!\nWARNING: No options to save results have been "
                    "specified.\nModel will run without saving results!\n!!!\n",
                    fg="red",
                    bold=True,
                )
            # If save_netcdf is used, override the 'save_per_spore_path' to point to a
            # directory of the same name as the planned netcdf

            if save_netcdf and model.run_config.get("spores_options", {}).get(
                "save_per_spore", False
            ):
                model.run_config["spores_options"][
                    "save_per_spore_path"
                ] = save_netcdf.replace(".nc", "/spore_{}.nc")

            model.run()
            termination = model._model_data.attrs.get(
                "termination_condition", "unknown"
            )

            if save_csv:
                click.secho("Saving CSV results to directory: {}".format(save_csv))
                model.to_csv(save_csv)
            if save_netcdf:
                click.secho("Saving NetCDF results to file: {}".format(save_netcdf))
                model.to_netcdf(save_netcdf)
            if save_plots:
                if termination == "optimal":
                    click.secho("Saving HTML file with plots to: {}".format(save_plots))
                    model.plot.summary(to_file=save_plots)
                else:
                    click.secho(
                        "Model termination condition non-optimal. Not saving plots",
                        fg="red",
                        bold=True,
                    )

            print_end_time(start_time)
            if fail_when_infeasible and termination != "optimal":
                raise BackendError("Problem is infeasible.")


@cli.command(
    name="generate_runs", short_help="Generate a script to run multiple models."
)
@click.argument("model_file")
@click.argument("out_file")
@click.option("--kind", help='One of: "bash", "bsub", "sbatch", or "windows".')
@click.option("--scenarios")
@click.option("--cluster_threads", default=1)
@click.option("--cluster_mem")
@click.option("--cluster_time")
@click.option(
    "--additional_args",
    default="",
    help="Any additional arguments to pass directly on to `calliope run`.",
)
@click.option("--override_dict")
@_debug
@_quiet
@_pdb
def generate_runs(
    model_file,
    out_file,
    kind,
    scenarios,
    additional_args,
    override_dict,
    cluster_threads,
    cluster_mem,
    cluster_time,
    debug,
    quiet,
    pdb,
):

    _cli_start(debug, quiet)

    kwargs = dict(
        model_file=model_file,
        out_file=out_file,
        scenarios=scenarios,
        additional_args=additional_args,
        override_dict=override_dict,
        cluster_mem=cluster_mem,
        cluster_time=cluster_time,
        cluster_threads=cluster_threads,
    )

    with format_exceptions(debug, pdb):
        generate(kind, **kwargs)


@cli.command(
    name="generate_scenarios",
    short_help="Generate scenario definitions from given combinations of overrides.",
)
@click.argument("model_file")
@click.argument("out_file")
@click.argument("overrides", nargs=-1)
@click.option("--scenario_name_prefix")
@_debug
@_quiet
@_pdb
def generate_scenarios(
    model_file, out_file, overrides, scenario_name_prefix, debug, quiet, pdb
):

    _cli_start(debug, quiet)

    with format_exceptions(debug, pdb):
        combinations = list(itertools.product(*[i.split(";") for i in overrides]))

        if not scenario_name_prefix:
            scenario_name_prefix = "scenario_"

        # len(str(x)) gives us the number of digits in x, for padding
        scenario_string = "{}{:0>" + str(len(str(len(combinations)))) + "d}"

        scenarios = {
            "scenarios": {
                scenario_string.format(scenario_name_prefix, i + 1): c
                for i, c in enumerate(combinations)
            }
        }

        AttrDict(scenarios).to_yaml(out_file)



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

policy.py
~~~~~~~~~

Policy constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

ORDER = 10  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]
    run_config = backend_model.__calliope_run_config

    if "techlists_group_share_energy_cap_min_constraint" in sets:
        backend_model.group_share_energy_cap_min_constraint = po.Constraint(
            backend_model.techlists_group_share_energy_cap_min_constraint,
            ["min"],
            rule=group_share_energy_cap_constraint_rule,
        )

    if "techlists_group_share_energy_cap_max_constraint" in sets:
        backend_model.group_share_energy_cap_max_constraint = po.Constraint(
            backend_model.techlists_group_share_energy_cap_max_constraint,
            ["max"],
            rule=group_share_energy_cap_constraint_rule,
        )

    if "techlists_group_share_energy_cap_equals_constraint" in sets:
        backend_model.group_share_energy_cap_equals_constraint = po.Constraint(
            backend_model.techlists_group_share_energy_cap_equals_constraint,
            ["equals"],
            rule=group_share_energy_cap_constraint_rule,
        )

    if "techlists_carrier_group_share_carrier_prod_min_constraint" in sets:
        backend_model.group_share_carrier_prod_min_constraint = po.Constraint(
            backend_model.techlists_carrier_group_share_carrier_prod_min_constraint,
            ["min"],
            rule=group_share_carrier_prod_constraint_rule,
        )

    if "techlists_carrier_group_share_carrier_prod_max_constraint" in sets:
        backend_model.group_share_carrier_prod_max_constraint = po.Constraint(
            backend_model.techlists_carrier_group_share_carrier_prod_max_constraint,
            ["max"],
            rule=group_share_carrier_prod_constraint_rule,
        )

    if "techlists_carrier_group_share_carrier_prod_equals_constraint" in sets:
        backend_model.group_share_carrier_prod_equals_constraint = po.Constraint(
            backend_model.techlists_carrier_group_share_carrier_prod_equals_constraint,
            ["equals"],
            rule=group_share_carrier_prod_constraint_rule,
        )

    if "carriers_reserve_margin_constraint" in sets and run_config["mode"] != "operate":
        backend_model.reserve_margin_constraint = po.Constraint(
            backend_model.carriers_reserve_margin_constraint,
            rule=reserve_margin_constraint_rule,
        )


def equalizer(lhs, rhs, sign):
    if sign == "max":
        return lhs <= rhs
    elif sign == "min":
        return lhs >= rhs
    elif sign == "equals":
        return lhs == rhs
    else:
        raise ValueError("Invalid sign: {}".format(sign))


# FIXME: docstring should show all variants: \geq, =, and \leq
def group_share_energy_cap_constraint_rule(backend_model, techlist, what):
    """
    Enforce shares in energy_cap for groups of technologies. Applied
    to ``supply`` and ``supply_plus`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech \\in given\\_group} energy_{cap}(loc::tech) =
            fraction \\times \\sum_{loc::tech \\in loc\\_techs\\_supply \\or loc\\_techs\\_supply\\_plus} energy_{cap}(loc::tech)

    """
    model_data_dict = backend_model.__calliope_model_data["data"]
    sets = backend_model.__calliope_model_data["sets"]
    fraction = model_data_dict["group_share_energy_cap_{}".format(what)][techlist]

    if "loc_techs_supply_plus" in sets:
        rhs_loc_techs = (
            backend_model.loc_techs_supply | backend_model.loc_techs_supply_plus
        )
        lhs_loc_techs = [
            i
            for i in backend_model.loc_techs_supply
            | backend_model.loc_techs_supply_plus
            if i.split("::")[1] in techlist.split(",")
        ]
    else:
        rhs_loc_techs = backend_model.loc_techs_supply
        lhs_loc_techs = [
            i
            for i in backend_model.loc_techs_supply
            if i.split("::")[1] in techlist.split(",")
        ]

    rhs = fraction * sum(
        backend_model.energy_cap[loc_tech] for loc_tech in rhs_loc_techs
    )
    lhs = sum(backend_model.energy_cap[loc_tech] for loc_tech in lhs_loc_techs)

    return equalizer(lhs, rhs, what)


def group_share_carrier_prod_constraint_rule(backend_model, techlist_carrier, what):
    """
    Enforce shares in carrier_prod for groups of technologies. Applied
    to ``loc_tech_carriers_supply_conversion_all``, which includes supply,
    supply_plus, conversion, and conversion_plus.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group, timestep \\in timesteps} carrier_{prod}(loc::tech::carrier, timestep) =
            fraction \\times \\sum_{loc::tech:carrier \\in loc\\_tech\\_carriers\\_supply\\_all, timestep\\in timesteps} carrier_{prod}(loc::tech::carrier, timestep)

    """
    model_data_dict = backend_model.__calliope_model_data["data"]
    techlist, carrier = techlist_carrier.split("::")
    fraction = model_data_dict["group_share_carrier_prod_{}".format(what)][
        (carrier, techlist)
    ]

    rhs_loc_tech_carriers = [
        i
        for i in backend_model.loc_tech_carriers_supply_conversion_all
        if i.split("::")[-1] == carrier
    ]
    lhs_loc_tech_carriers = [
        i
        for i in backend_model.loc_tech_carriers_supply_conversion_all
        if i.split("::")[1] in techlist.split(",") and i.split("::")[-1] == carrier
    ]
    rhs = fraction * sum(
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        for loc_tech_carrier in rhs_loc_tech_carriers
        for timestep in backend_model.timesteps
    )
    lhs = sum(
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        for loc_tech_carrier in lhs_loc_tech_carriers
        for timestep in backend_model.timesteps
    )

    return equalizer(lhs, rhs, what)


def reserve_margin_constraint_rule(backend_model, carrier):
    """
    Enforces a system reserve margin per carrier.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc\\_tech\\_carriers\\_supply\\_all} energy_{cap}(loc::tech::carrier, timestep_{max\\_demand})
            \\geq
            \\sum_{loc::tech::carrier \\in loc\\_tech\\_carriers\\_demand} carrier_{con}(loc::tech::carrier, timestep_{max\\_demand})
            \\times -1 \\times \\frac{1}{time\\_resolution_{max\\_demand}}
            \\times (1 + reserve\\_margin)

    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    reserve_margin = model_data_dict["reserve_margin"][carrier]
    max_demand_timestep = model_data_dict["max_demand_timesteps"][carrier]
    max_demand_time_res = backend_model.timestep_resolution[max_demand_timestep]

    return sum(  # Sum all supply capacity for this carrier
        backend_model.energy_cap[loc_tech_carrier.rsplit("::", 1)[0]]
        for loc_tech_carrier in backend_model.loc_tech_carriers_supply_conversion_all
        if loc_tech_carrier.split("::")[-1] == carrier
    ) >= sum(  # Sum all demand for this carrier and timestep
        backend_model.carrier_con[loc_tech_carrier, max_demand_timestep]
        for loc_tech_carrier in backend_model.loc_tech_carriers_demand
        if loc_tech_carrier.split("::")[-1] == carrier
    ) * -1 * (
        1 / max_demand_time_res
    ) * (
        1 + reserve_margin
    )



from calliope.preprocess.model_run import (
    model_run_from_yaml,
    model_run_from_dict,
    generate_model_run,
    apply_overrides,
)

from calliope.preprocess.model_data import build_model_data

from calliope.preprocess.time import (
    apply_time_clustering,
    final_timedimension_processing,
)



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

timeseries.py
~~~~~~~~~~~~~

Plot timeseries data.

"""

import pandas as pd
import numpy as np
import plotly.graph_objs as go

from calliope.postprocess.util import subset_sum_squeeze
from calliope.postprocess.plotting.util import (
    get_data_layout,
    hex_to_rgba,
    break_name,
    get_clustered_layout,
)


def _get_relevant_vars(model, dataset, array):

    carriers = list(dataset.carriers.values)

    allowed_input_vars = [
        k
        for k, v in model.inputs.data_vars.items()
        if "timesteps" in v.dims and len(v.dims) > 1 and k != "carrier_ratios"
    ]
    allowed_result_vars = [
        "results",
        "inputs",
        "all",
        "storage",
        "storage_inter_cluster",
        "resource_con",
        "cost_var",
    ]

    if (
        isinstance(array, list)
        and not set(array).intersection(
            allowed_input_vars + allowed_result_vars + carriers
        )
    ) or (
        isinstance(array, str)
        and array not in allowed_input_vars + allowed_result_vars + carriers
    ):
        raise ValueError(
            "Cannot plot array={}. If you want carrier flow (_prod, _con, _export) "
            "then specify the name of the energy carrier as array".format(array)
        )

    # relevant_vars are all variables relevant to this plotting instance
    relevant_vars = []

    # Ensure carriers are at the top of the list
    if array == "results":
        relevant_vars += sorted(carriers) + sorted(allowed_result_vars)
    elif array == "inputs" or (array == "all" and not hasattr(model, "results")):
        relevant_vars += sorted(allowed_input_vars)
    elif array == "all":
        relevant_vars += sorted(carriers) + sorted(
            allowed_result_vars + allowed_input_vars
        )
    elif isinstance(array, list):
        relevant_vars = array
    elif isinstance(array, str):
        relevant_vars = [array]

    relevant_vars = [i for i in relevant_vars if i in dataset or i in carriers]
    return relevant_vars


def _get_var_data(var, model, dataset, visible, subset, sum_dims, squeeze):
    """
    Get variable data from model_data and use it to populate a list with Plotly plots
    """
    carriers = list(dataset.carriers.values)

    # list to populate
    data = []

    # Clustered data does not get plotted on a datetime axis
    if "clusters" in dataset.dims:
        clusters = dataset.timestep_cluster.to_pandas()
        cluster_groups = clusters.groupby(clusters).groups
        # Add an extra timestep at the end of each cluster, so we can break
        # scatter lines with a NaN between each cluster
        dummy_timestep = lambda x: pd.to_datetime(x[-1] + pd.Timedelta(100, unit="ns"))
        reindexing_timesteps = np.concatenate(
            [
                group_timesteps.append(
                    pd.Index([dummy_timestep(group_timesteps)])
                ).values
                for group_timesteps in cluster_groups.values()
            ]
        )
        timesteps = pd.Index(reindexing_timesteps).values.astype(str)
        # replace dummy timestep with NaN
        timesteps[pd.Index(reindexing_timesteps).nanosecond == 100] = np.nan
        timesteps = pd.to_datetime(timesteps, errors="ignore")
        if var == "storage_inter_cluster":
            datesteps = pd.to_datetime(model._model_data.datesteps.values)
    else:
        timesteps = pd.to_datetime(model._model_data.timesteps.values)

    def _get_reindexed_array(array, index=["locs", "techs"], fillna=None):
        # reindexing data means that DataArrays have the same values in locs and techs
        reindexer = {k: sorted(dataset[k].values) for k in index}
        formatted_array = model.get_formatted_array(array)
        if fillna is not None:
            return formatted_array.reindex(**reindexer).fillna(fillna)
        else:
            return formatted_array.reindex(**reindexer)

    def _get_y_vals(array, ignore_clustering=False):
        if "clusters" in dataset.dims and not ignore_clustering:
            return array.to_pandas().reindex(reindexing_timesteps).values
        else:
            return array.values

    if hasattr(model, "results") and len(model.results.data_vars) > 0:
        array_prod = _get_reindexed_array(
            "carrier_prod", index=["locs", "techs", "carriers"], fillna=0
        )
        array_con = _get_reindexed_array(
            "carrier_con", index=["locs", "techs", "carriers"], fillna=0
        )
        if "resource_con" in dataset.data_vars:
            resource_con = _get_reindexed_array("resource_con", fillna=0)
        else:
            resource_con = 0

    # carrier flow is a combination of carrier_prod, carrier_con and
    # carrier_export for a given energy carrier
    if var in carriers:
        array_flow = (
            array_prod.loc[dict(carriers=var)] + array_con.loc[dict(carriers=var)]
        )
        if "carrier_export" in dataset:
            export_flow = subset_sum_squeeze(
                _get_reindexed_array(
                    "carrier_export", index=["locs", "techs", "carriers"], fillna=0
                ).loc[dict(carriers=var)],
                subset,
                sum_dims,
                squeeze,
            )
        if "unmet_demand" in dataset:
            unmet_flow = subset_sum_squeeze(
                _get_reindexed_array(
                    "unmet_demand", index=["locs", "carriers"], fillna=0
                ).loc[dict(carriers=var)],
                subset,
                sum_dims,
                squeeze=False,
            )

    # array flow for storage tracks stored energy. carrier_flow is
    # charge/discharge (including resource consumed for supply_plus techs)
    elif var == "storage":
        array_flow = _get_reindexed_array("storage")

        if "resource_eff" in dataset.data_vars:
            resource_eff = _get_reindexed_array("resource_eff", fillna=1)
        else:
            resource_eff = 1

        charge = subset_sum_squeeze(
            -array_con.sum("carriers") + resource_con * resource_eff,
            subset,
            sum_dims,
            squeeze=False,
        )
        discharge = -subset_sum_squeeze(
            array_prod.sum("carriers"), subset, sum_dims, squeeze=False
        )

    elif var == "resource_con":
        array_flow = resource_con

    else:
        array_flow = _get_reindexed_array(var)

    array_flow = subset_sum_squeeze(array_flow, subset, sum_dims, squeeze)

    err_details = " (variable: `{}`, subset: `{}`, sum_dims: `{}`.".format(
        var, subset, sum_dims
    )

    if "timesteps" not in array_flow.dims and "datesteps" not in array_flow.dims:
        raise ValueError(
            "`timesteps` not in plotting data, cannot proceed" + err_details
        )

    if len(array_flow.dims) > 2:
        if "costs" in array_flow.dims:
            err_details = err_details + (
                " Try subsetting to select a cost class, "
                "e.g. subset={'costs': ['monetary']}."
            )
        raise ValueError(
            "Too many dimensions to plot: `{}`".format(array_flow.dims) + err_details
        )

    # If techs not given as a subset (implying a desire to order manually),
    # sort techs such that those varying output least are at the bottom of the stack
    if subset.get("techs", None) or var == "storage_inter_cluster":
        sorted_techs = array_flow.techs.values
    else:
        sorted_techs = [
            array_flow.techs.values[i]
            for i in array_flow.var(dim="timesteps").argsort()
        ]

    # for tech in array_flow.techs.values:
    for tech in sorted_techs:
        tech_dict = {"techs": tech}
        if not array_flow.loc[tech_dict].sum():
            continue
        # We allow transmission tech information to show up in some cases
        if (
            "techs_transmission" in dataset
            and tech in dataset.techs_transmission.values
        ):
            base_tech = "transmission"
            color = dataset.colors.loc[{"techs": tech.split(":")[0]}].item()
            name = break_name(
                dataset.names.loc[{"techs": tech.split(":")[0]}].item(), 30
            )
            if var in carriers:
                continue  # no transmission in carrier flow
        else:
            base_tech = dataset.inheritance.loc[tech_dict].item().split(".")[0]
            color = dataset.colors.loc[tech_dict].item()
            name = break_name(dataset.names.loc[tech_dict].item(), 30)

        if base_tech == "demand":
            # Always insert demand at position 0 in the list, to make
            # sure it appears on top in the legend
            data.insert(
                0,
                go.Scatter(
                    x=timesteps,
                    y=_get_y_vals(-array_flow.loc[tech_dict]),
                    visible=visible,
                    line=dict(color=color),
                    name=name,
                    legendgroup=tech,
                ),
            )

        elif var == "storage":
            # stored energy as scatter, carrier/resource prod/con as stacked bar
            data.insert(
                0,
                go.Scatter(
                    x=timesteps,
                    y=_get_y_vals(array_flow.loc[tech_dict]),
                    visible=visible,
                    line=dict(color=color),
                    mode="lines",
                    name=name + " stored energy",
                    showlegend=False,
                    text=tech + " stored energy",
                    hoverinfo="x+y+text",
                    legendgroup=tech,
                ),
            )
            data.append(
                go.Bar(
                    x=timesteps,
                    y=_get_y_vals(charge.loc[tech_dict]),
                    visible=visible,
                    name=name,
                    marker=dict(color=color),
                    legendgroup=tech,
                    text=tech + " charge",
                    hoverinfo="x+y+text",
                )
            )
            data.append(
                go.Bar(
                    x=timesteps,
                    y=_get_y_vals(discharge.loc[tech_dict]),
                    visible=visible,
                    name=name,
                    marker=dict(color=color, opacity=0.8),
                    legendgroup=tech,
                    text=tech + " discharge",
                    hoverinfo="x+y+text",
                    showlegend=False,
                )
            )

        elif var == "storage_inter_cluster":
            data.append(
                go.Scatter(
                    x=datesteps,
                    y=_get_y_vals(array_flow.loc[tech_dict], ignore_clustering=True),
                    visible=visible,
                    line=dict(color=color),
                    mode="lines",
                    name=name,
                    showlegend=False,
                    text=tech,
                    hoverinfo="x+y+text",
                    legendgroup=tech,
                )
            )

        else:
            data.append(
                go.Bar(
                    x=timesteps,
                    y=_get_y_vals(array_flow.loc[tech_dict]),
                    visible=visible,
                    name=name,
                    legendgroup=tech,
                    marker=dict(color=color),
                )
            )

        if (
            var in carriers
            and "carrier_export" in dataset
            and export_flow.loc[tech_dict].sum()
        ):
            data.append(
                go.Bar(
                    x=timesteps,
                    y=_get_y_vals(-export_flow.loc[tech_dict]),
                    visible=visible,
                    name=name + " export",
                    legendgroup=tech,
                    marker=dict(color=hex_to_rgba(color, 0.5)),
                )
            )

    if var in carriers and "unmet_demand" in dataset:
        data.append(
            go.Bar(
                x=timesteps,
                y=_get_y_vals(unmet_flow),
                visible=visible,
                name="Unmet " + var + " demand",
                legendgroup="unmet_demand",
                marker=dict(color="grey"),
            )
        )

    return data


def _get_var_layout(var, dataset):
    """
    Variable-specific layout. Increases axis verbosity for some known variables.
    `visible` used in dropdown, not if only one array is shown.

    """
    args = {}
    if var in dataset.carriers.values:
        title = "Carrier flow: {}".format(var)
        y_axis_title = "Energy produced(+) / consumed(-)"
    elif var == "resource":
        title = "Available resource"
        y_axis_title = "Energy (per unit of area)"
    elif var == "resource_con":
        title = "Consumed resource"
        y_axis_title = "Energy"
    elif var == "cost_var":
        title = "Variable costs"
        y_axis_title = "Cost"
    elif var == "storage_inter_cluster":
        title = "Storage at start of each day in unclustered timeseries"
        y_axis_title = "Stored energy"
        args.update({"xaxis": {"type": "date"}})
    else:
        title = y_axis_title = "{}".format(var).capitalize()

    if "clusters" in dataset.dims and var != "storage_inter_cluster":
        args.update(get_clustered_layout(dataset))

    args.update({"yaxis": dict(title=y_axis_title), "title": title})

    return args


def plot_timeseries(
    model,
    array="all",
    timesteps_zoom=None,
    rangeslider=False,
    subset={},
    sum_dims="locs",
    squeeze=True,
    **kwargs,
):
    """
    Parameters
    ----------
    array : str or list; default = 'all'
        options: 'all', 'results', 'inputs', the name/list of any energy carrier(s)
        (e.g. 'power'), the name/list of any input/output DataArray(s).

        User can specify 'all' for all input/results timeseries plots, 'inputs'
        for just input timeseries, 'results' for just results timeseries, or the
        name of any data array to plot (in either inputs or results).
        In all but the last case, arrays can be picked from dropdown in visualisation.
        In the last case, output can be saved to SVG and a rangeslider can be used.

    timesteps_zoom : int, optional
        Number of timesteps to show initially on the x-axis (if not
        given, the full time range is shown by default).
    rangeslider : bool, optional
        If True, displays a range slider underneath the plot for navigating
        (helpful primarily in interactive use).
    subset : dict, optional
        Dictionary by which data is subset (uses xarray `loc` indexing). Keys
        any of ['timeseries', 'locs', 'techs', 'carriers', 'costs'].
    sum_dims : str, optional
        List of dimension names to sum plot variable over.
    squeeze : bool, optional
        Whether to squeeze out dimensions of length = 1.

    """

    dataset = model._model_data.copy()

    timesteps = pd.to_datetime(model._model_data.timesteps.values)

    layout = dict(
        barmode="relative",
        xaxis={},
        autosize=True,
        legend=(dict(traceorder="reversed", xanchor="left")),
        hovermode="x",
    )
    # Clustered data does not get plotted on a datetime axis
    if "clusters" in dataset.dims:
        layout.update(get_clustered_layout(dataset))

    relevant_vars = _get_relevant_vars(model, dataset, array)

    data, layout = get_data_layout(
        _get_var_data,
        _get_var_layout,
        relevant_vars,
        layout,
        model,
        dataset,
        subset,
        sum_dims,
        squeeze,
    )

    if rangeslider:
        layout["xaxis"]["rangeslider"] = {}

    if timesteps_zoom:
        layout["xaxis"]["range"] = [timesteps[0], timesteps[timesteps_zoom]]

    return data, layout
